import os
import warnings
from pathlib import Path
from xml.etree import ElementTree

import dask
import dask.array as da
import numpy as np
import pandas as pd
import xarray as xr

from dtscalibration.io.utils import dim_attrs
from dtscalibration.io.utils import get_xml_namespace
from dtscalibration.io.utils import open_file

dim_attrs_apsensing = dict(dim_attrs)
dim_attrs_apsensing["TEMP"] = dim_attrs_apsensing.pop("tmp")
dim_attrs_apsensing["TEMP"]["name"] = "TEMP"
dim_attrs_apsensing.pop("acquisitionTime")
dim_attrs_apsensing.pop("userAcquisitionTimeFW")
dim_attrs_apsensing.pop("userAcquisitionTimeBW")


def read_apsensing_files(
    filepathlist=None,
    directory=None,
    file_ext="*.xml",
    timezone_input_files="UTC",
    timezone_netcdf="UTC",
    silent=False,
    load_in_memory="auto",
    **kwargs,
):
    """Read a folder with measurement files from a device of the Sensortran
    brand. Each measurement file contains values for a single timestep.
    Remember to check which timezone you are working in.

    Parameters
    ----------
    filepathlist : list of str, optional
        List of paths that point the the silixa files
    directory : str, Path, optional
        Path to folder
    timezone_input_files : str, optional
        Timezone string of the measurement files.
        Remember to check when measurements are taken.
        Also if summertime is used.
    timezone_netcdf : str, optional
        Timezone string of the netcdf file. UTC follows CF-conventions.
    file_ext : str, optional
        file extension of the measurement files
    silent : bool
        If set tot True, some verbose texts are not printed to stdout/screen
    load_in_memory : {'auto', True, False}
        If 'auto' the Stokes data is only loaded to memory for small files
    kwargs : dict-like, optional
        keyword-arguments are passed to DataStore initialization

    Notes:
    ------
    Only XML files are supported for now

    Returns:
    --------
    datastore : DataStore
        The newly created datastore.
    """
    if not file_ext == "*.xml":
        raise NotImplementedError("Only .xml files are supported for now")

    if filepathlist is None:
        filepathlist = sorted(Path(directory).glob(file_ext))

        # Make sure that the list of files contains any files
        assert len(filepathlist) >= 1, (
            "No measurement files found in provided " "directory: \n" + str(directory)
        )

    # Make sure that the list of files contains any files
    assert len(filepathlist) >= 1, (
        "No measurement files found in provided " "list/directory"
    )

    device = apsensing_xml_version_check(filepathlist)

    valid_devices = ["CP320"]

    if device in valid_devices:
        pass

    else:
        warnings.warn(
            "AP sensing device {device}"
            " has not been tested.\nPlease open an issue on github"
            " and provide an example file"
        )

    data_vars, coords, attrs = read_apsensing_files_routine(
        filepathlist,
        timezone_netcdf=timezone_netcdf,
        timezone_input_files=timezone_input_files,
        silent=silent,
        load_in_memory=load_in_memory,
    )

    ds = xr.Dataset(data_vars=data_vars, coords=coords, attrs=attrs, **kwargs)
    return ds


def apsensing_xml_version_check(filepathlist):
    """Function which tests which version of xml files are read.

    Parameters
    ----------
    filepathlist

    Returns:
    --------

    """
    sep = ":"
    attrs, _ = read_apsensing_attrs_singlefile(filepathlist[0], sep)

    return attrs["wellbore:uid"]


def read_apsensing_files_routine(
    filepathlist,
    timezone_input_files="UTC",
    timezone_netcdf="UTC",
    silent=False,
    load_in_memory="auto",
):
    """Internal routine that reads AP Sensing files.
    Use dtscalibration.read_apsensing_files function instead.

    The AP sensing files are not timezone aware

    Parameters
    ----------
    filepathlist
    timezone_input_files
    timezone_netcdf
    silent
    load_in_memory

    Returns:
    --------

    """
    assert (
        timezone_input_files == "UTC" and timezone_netcdf == "UTC"
    ), "Only UTC timezones supported"

    # translate names
    tld = {"ST": "st", "AST": "ast", "REV-ST": "rst", "REV-AST": "rast", "TEMP": "tmp"}

    # Open the first xml file using ET, get the name space and amount of data
    xml_tree = ElementTree.parse(filepathlist[0])
    namespace = get_xml_namespace(xml_tree.getroot())

    logtree = xml_tree.find(
        (
            "{0}wellSet/{0}well/{0}wellboreSet/{0}wellbore"
            + "/{0}wellLogSet/{0}wellLog"
        ).format(namespace)
    )
    logdata_tree = logtree.find(f"./{namespace}logData")

    # Amount of datapoints is the size of the logdata tree
    nx = len(logdata_tree)

    sep = ":"
    ns = {"s": namespace[1:-1]}

    # Obtain metadata from the first file
    attrs, skip_chars = read_apsensing_attrs_singlefile(filepathlist[0], sep)

    # Add standardised required attributes
    # No example of DE file available
    attrs["isDoubleEnded"] = "0"
    double_ended_flag = bool(int(attrs["isDoubleEnded"]))

    attrs["forwardMeasurementChannel"] = attrs[
        "wellbore:dtsMeasurementSet:dtsMeasurement:connectedToFiber:uidRef"
    ]
    attrs["backwardMeasurementChannel"] = "N/A"

    data_item_names = [
        attrs[f"wellbore:wellLogSet:wellLog:logCurveInfo_{x}:mnemonic"]
        for x in range(0, 4)
    ]

    nitem = len(data_item_names)

    ntime = len(filepathlist)

    # print summary
    if not silent:
        print("%s files were found, each representing a single timestep" % ntime)
        print("%s recorded vars were found: " % nitem + ", ".join(data_item_names))
        print("Recorded at %s points along the cable" % nx)

        if double_ended_flag:
            print("The measurement is double ended")
        else:
            print("The measurement is single ended")

    # Gather data
    arr_path = "s:" + "/s:".join(
        [
            "wellSet",
            "well",
            "wellboreSet",
            "wellbore",
            "wellLogSet",
            "wellLog",
            "logData",
            "data",
        ]
    )

    @dask.delayed
    def grab_data_per_file(file_handle):
        """Parameters
        ----------
        file_handle

        Returns:
        --------

        """
        with open_file(file_handle, mode="r") as f_h:
            if skip_chars:
                f_h.read(3)
            eltree = ElementTree.parse(f_h)
            arr_el = eltree.findall(arr_path, namespaces=ns)

            if not len(arr_el) == nx:
                raise ValueError(
                    "Inconsistent length of x-dimension"
                    + "\nCheck if files are mixed up, or if the number of "
                    + "data points vary per file."
                )

            # remove the breaks on both sides of the string
            # split the string on the comma
            arr_str = [arr_eli.text.split(",") for arr_eli in arr_el]
        return np.array(arr_str, dtype=float)

    data_lst_dly = [grab_data_per_file(fp) for fp in filepathlist]

    data_lst = [
        da.from_delayed(x, shape=(nx, nitem), dtype=float) for x in data_lst_dly
    ]
    data_arr = da.stack(data_lst).T  # .compute()

    # Check whether to compute data_arr (if possible 25% faster)
    data_arr_cnk = data_arr.rechunk({0: -1, 1: -1, 2: "auto"})
    if load_in_memory == "auto" and data_arr_cnk.npartitions <= 5 or load_in_memory:
        if not silent:
            print("Reading the data from disk")
        data_arr = data_arr_cnk.compute()
    else:
        if not silent:
            print("Not reading the data from disk")
        data_arr = data_arr_cnk

    data_vars = {}
    for name, data_arri in zip(data_item_names, data_arr):
        if name == "LAF":
            continue

        if tld[name] in dim_attrs_apsensing:
            data_vars[tld[name]] = (
                ["x", "time"],
                data_arri,
                dim_attrs_apsensing[tld[name]],
            )
        elif name in dim_attrs_apsensing:
            data_vars[tld[name]] = (["x", "time"], data_arri, dim_attrs_apsensing[name])
        else:
            raise ValueError("Dont know what to do with the" + f" {name} data column")

    _time_dtype = [("filename_tstamp", np.int64), ("acquisitionTime", "<U29")]
    ts_dtype = np.dtype(_time_dtype)

    @dask.delayed
    def grab_timeseries_per_file(file_handle):
        """Parameters
        ----------
        file_handle

        Returns:
        --------

        """
        with open_file(file_handle, mode="r") as f_h:
            if skip_chars:
                f_h.read(3)
            eltree = ElementTree.parse(f_h)

            out = []

            # get all the time related data
            creationDate = eltree.find(
                (
                    "{0}wellSet/{0}well/{0}wellboreSet"
                    + "/{0}wellbore/{0}wellLogSet"
                    + "/{0}wellLog/{0}creationDate"
                ).format(namespace)
            ).text

            if isinstance(file_handle, tuple):
                file_name = os.path.split(file_handle[0])[-1]
            else:
                file_name = os.path.split(file_handle)[-1]

            tstamp = np.int64(file_name[-20:-4])

            out += [tstamp, creationDate]
        return np.array(tuple(out), dtype=ts_dtype)

    ts_lst_dly = [grab_timeseries_per_file(fp) for fp in filepathlist]
    ts_lst = [da.from_delayed(x, shape=tuple(), dtype=ts_dtype) for x in ts_lst_dly]
    ts_arr = da.stack(ts_lst).compute()

    data_vars["creationDate"] = (
        ("time",),
        [pd.Timestamp(str(item[1])) for item in ts_arr],
    )

    # construct the coordinate dictionary
    coords = {
        "x": ("x", data_arr[0, :, 0], dim_attrs_apsensing["x"]),
        "filename": ("time", [os.path.split(f)[1] for f in filepathlist]),
        "time": data_vars["creationDate"],
    }

    return data_vars, coords, attrs


def read_apsensing_attrs_singlefile(filename, sep):
    """Parameters
    ----------
    filename
    sep

    Returns:
    --------

    """
    from xml.parsers.expat import ExpatError

    import xmltodict

    def metakey(meta, dict_to_parse, prefix):
        """Fills the metadata dictionairy with data from dict_to_parse.
        The dict_to_parse is the raw data from a silixa xml-file.
        dict_to_parse is a nested dictionary to represent the
        different levels of hierarchy. For example,
        toplevel = {lowlevel: {key: value}}.
        This function returns {'toplevel:lowlevel:key': value}.
        Where prefix is the flattened hierarchy.

        Parameters
        ----------
        meta : dict
            the output dictionairy with prcessed metadata
        dict_to_parse : dict
        prefix

        Returns:
        --------

        """
        for key in dict_to_parse:
            if prefix == "":
                prefix_parse = key.replace("@", "")
            else:
                prefix_parse = sep.join([prefix, key.replace("@", "")])

            items = ("wellbore", "wellLogSet", "wellLog", "logData", "data")
            if prefix_parse == sep.join(items):
                continue

            if hasattr(dict_to_parse[key], "keys"):
                # Nested dictionaries, flatten hierarchy.
                meta.update(metakey(meta, dict_to_parse[key], prefix_parse))

            elif isinstance(dict_to_parse[key], list):
                # if the key has values for the multiple channels
                for ival, val in enumerate(dict_to_parse[key]):
                    num_key = prefix_parse + "_" + str(ival)
                    meta.update(metakey(meta, val, num_key))
            else:
                meta[prefix_parse] = dict_to_parse[key]

        return meta

    with open_file(filename) as fh:
        data = fh.read()
        try:
            doc_ = xmltodict.parse(data)
            skip_chars = False
        # the first 3 characters can be weird, skip them
        except ExpatError:
            doc_ = xmltodict.parse(data[3:])
            skip_chars = True

    doc = doc_["WITSMLComposite"]["wellSet"]["well"]["wellboreSet"]

    return metakey({}, doc, ""), skip_chars



import fnmatch
import os
import warnings
from glob import glob

import numpy as np
import pandas as pd
import xarray as xr

from dtscalibration.io.utils import coords_time
from dtscalibration.io.utils import dim_attrs
from dtscalibration.io.utils import open_file


def read_sensornet_files(
    filepathlist=None,
    directory=None,
    file_ext="*.ddf",
    timezone_input_files="UTC",
    timezone_netcdf="UTC",
    silent=False,
    add_internal_fiber_length=50.0,
    fiber_length=None,
    **kwargs,
):
    """Read a folder with measurement files. Each measurement file contains
    values for a single timestep. Remember to check which timezone
    you are working in.

    Parameters
    ----------
    filepathlist : list of str, optional
        List of paths that point the the silixa files
    directory : str, Path, optional
        Path to folder
    timezone_input_files : str, optional
        Timezone string of the measurement files.
        Remember to check when measurements are taken.
        Also if summertime is used.
    timezone_netcdf : str, optional
        Timezone string of the netcdf file. UTC follows CF-conventions.
    file_ext : str, optional
        file extension of the measurement files
    silent : bool
        If set tot True, some verbose texts are not printed to stdout/screen
    add_internal_fiber_length : float
        Set to zero if only the measurements of the fiber connected to the DTS
        system of interest. Set to 50 if you also want to keep the internal
        reference section.
    fiber_length : float
        It is the fiber length between the two connector entering the DTS
        device. If left to `None`, it is approximated with
        `x[-1] - add_internal_fiber_length`.
    kwargs : dict-like, optional
        keyword-arguments are passed to DataStore initialization

    Notes:
    ------
    Compressed sensornet files can not be directly decoded
    because the files are encoded with encoding='windows-1252' instead of
    UTF-8.

    Returns:
    --------
    datastore : DataStore
        The newly created datastore.
    """
    if filepathlist is None:
        # Also look for files in sub-folders
        filepathlist_unsorted = glob(
            os.path.join(directory, "**", file_ext), recursive=True
        )

        # Make sure that the list of files contains any files
        msg = "No measurement files found in provided directory: \n" + str(directory)
        assert len(filepathlist_unsorted) >= 1, msg

        # sort based on dates in filesname. A simple sorted() is not sufficient
        # as month folders do not sort well
        basenames = [os.path.basename(fp) for fp in filepathlist_unsorted]
        dates = ["".join(bn.split(" ")[2:4]) for bn in basenames]
        i_sort = np.argsort(dates)
        filepathlist = [filepathlist_unsorted[i] for i in i_sort]

        # Check measurements are all from same channel
        chno = [bn.split(" ")[1] for bn in basenames]
        assert (
            len(set(chno)) == 1
        ), "Folder contains measurements from multiple channels"

    # Make sure that the list of files contains any files
    assert len(filepathlist) >= 1, (
        "No measurement files found in provided " "list/directory"
    )

    ddf_version = sensornet_ddf_version_check(filepathlist)

    valid_versions = [
        "Halo DTS v1*",
        "ORYX F/W v1.02 Oryx Data Collector v3*",
        "ORYX F/W v4.00 Oryx Data Collector v3*",
        "Sentinel DTS v5*",
    ]

    valid = any([fnmatch.fnmatch(ddf_version, v_) for v_ in valid_versions])

    if valid and (
        fnmatch.fnmatch(ddf_version, "Halo DTS v1*")
        or fnmatch.fnmatch(ddf_version, "Sentinel DTS v5*")
    ):
        flip_reverse_measurements = True
    elif fnmatch.fnmatch(ddf_version, "ORYX F/W v4*"):
        flip_reverse_measurements = False
    else:
        flip_reverse_measurements = False
        warnings.warn(
            f"\n    Sensornet .dff version {ddf_version}"
            " has not been tested.\n    Please open an issue on github"
            " and provide an example file"
        )

    data_vars, coords, attrs = read_sensornet_files_routine_v3(
        filepathlist,
        timezone_netcdf=timezone_netcdf,
        timezone_input_files=timezone_input_files,
        silent=silent,
        add_internal_fiber_length=add_internal_fiber_length,
        fiber_length=fiber_length,
        flip_reverse_measurements=flip_reverse_measurements,
    )

    ds = xr.Dataset(data_vars=data_vars, coords=coords, attrs=attrs, **kwargs)
    return ds


def sensornet_ddf_version_check(filepathlist):
    """Function which checks and returns the .ddf file version.

    Parameters
    ----------
    filepathlist

    Returns:
    --------
    ddf_version

    """
    # Obtain metadata fro mthe first file
    _, meta = read_sensornet_single(filepathlist[0])

    if "Software version number" in meta:
        version_string = meta["Software version number"]
    else:
        raise ValueError(
            "Software version number could not be detected in .ddf file"
            + "Either file is corrupted or not supported"
        )

    ddf_version = version_string.replace(",", ".")

    return ddf_version


def read_sensornet_single(filename):
    """Parameters
    ----------
    filename

    Returns:
    --------

    """
    headerlength = 26

    # The $\circ$ Celsius symbol is unreadable in utf8
    with open_file(filename, encoding="windows-1252") as fileobject:
        filelength = sum([1 for _ in fileobject])
    datalength = filelength - headerlength

    meta = {}
    with open_file(filename, encoding="windows-1252") as fileobject:
        for ii in range(0, 4):
            fileline = fileobject.readline().split(":\t")
            meta[fileline[0]] = fileline[1].replace("\n", "")

        for ii in range(4, headerlength - 1):
            fileline = fileobject.readline().split("\t")
            meta[fileline[0]] = fileline[1].replace("\n", "").replace(",", ".")

        # data_names =
        fileobject.readline().split("\t")

        if meta["differential loss correction"] == "single-ended":
            data = {
                "x": np.zeros(datalength),
                "tmp": np.zeros(datalength),
                "st": np.zeros(datalength),
                "ast": np.zeros(datalength),
            }

            for ii in range(0, datalength):
                fileline = fileobject.readline().replace(",", ".").split("\t")

                data["x"][ii] = float(fileline[0])
                data["tmp"][ii] = float(fileline[1])
                data["st"][ii] = float(fileline[2])
                data["ast"][ii] = float(fileline[3])

        elif meta["differential loss correction"] == "combined":
            data = {
                "x": np.zeros(datalength),
                "tmp": np.zeros(datalength),
                "st": np.zeros(datalength),
                "ast": np.zeros(datalength),
                "rst": np.zeros(datalength),
                "rast": np.zeros(datalength),
            }

            for ii in range(0, datalength):
                fileline = fileobject.readline().replace(",", ".").split("\t")

                data["x"][ii] = float(fileline[0])
                data["tmp"][ii] = float(fileline[1])
                data["st"][ii] = float(fileline[2])
                data["ast"][ii] = float(fileline[3])
                data["rst"][ii] = float(fileline[4])
                data["rast"][ii] = float(fileline[5])

        else:
            raise ValueError(
                'unknown differential loss correction: "'
                + meta["differential loss correction"]
                + '"'
            )

    meta["default loss term dB per km"] = meta["default loss term (dB/km)"]
    del meta["default loss term (dB/km)"]

    return data, meta


def read_sensornet_files_routine_v3(
    filepathlist,
    timezone_netcdf="UTC",
    timezone_input_files="UTC",
    silent=False,
    add_internal_fiber_length=50.0,
    fiber_length=None,
    flip_reverse_measurements=False,
):
    """Internal routine that reads Sensor files.
    Use dtscalibration.read_sensornet_files function instead.

    Parameters
    ----------
    filepathlist
    timezone_netcdf
    timezone_input_files
    silent
    add_internal_fiber_length : float
        Set to zero if only the measurements of the fiber connected to the DTS
        system of interest. Set to 50 if you also want to keep the internal
        reference section.
    fiber_length : float
        It is the fiber length between the two connector entering the DTS
        device.

    Returns:
    --------

    """
    # Obtain metadata from the first file
    data, meta = read_sensornet_single(filepathlist[0])

    # Pop keys from the meta dict which are variable over time
    popkeys = (
        "T ext. ref 1 (°C)",
        "T ext. ref 2 (°C)",
        "T internal ref (°C)",
        "date",
        "time",
        "gamma",
        "k internal",
        "k external",
    )
    [meta.pop(key) for key in popkeys]
    attrs = meta

    # Add standardised required attributes
    if meta["differential loss correction"] == "single-ended":
        attrs["isDoubleEnded"] = "0"
    elif meta["differential loss correction"] == "combined":
        attrs["isDoubleEnded"] = "1"

    double_ended_flag = bool(int(attrs["isDoubleEnded"]))

    attrs["forwardMeasurementChannel"] = meta["forward channel"][-1]
    if double_ended_flag:
        attrs["backwardMeasurementChannel"] = "N/A"
    else:
        attrs["backwardMeasurementChannel"] = meta["reverse channel"][-1]

    # obtain basic data info
    nx = data["x"].size

    ntime = len(filepathlist)

    # chFW = int(attrs['forwardMeasurementChannel']) - 1  # zero-based
    # if double_ended_flag:
    #     chBW = int(attrs['backwardMeasurementChannel']) - 1  # zero-based
    # else:
    #     # no backward channel is negative value. writes better to netcdf
    #     chBW = -1

    # print summary
    if not silent:
        print("%s files were found," % ntime + " each representing a single timestep")
        print("Recorded at %s points along the cable" % nx)

        if double_ended_flag:
            print("The measurement is double ended")
        else:
            print("The measurement is single ended")

    #   Gather data
    # x has already been read. should not change over time
    xraw = data["x"]

    # Define all variables
    referenceTemperature = np.zeros(ntime)
    probe1temperature = np.zeros(ntime)
    probe2temperature = np.zeros(ntime)
    gamma_ddf = np.zeros(ntime)
    k_internal = np.zeros(ntime)
    k_external = np.zeros(ntime)
    acquisitiontimeFW = np.zeros(ntime)
    acquisitiontimeBW = np.zeros(ntime)

    timestamp = [""] * ntime
    ST = np.zeros((nx, ntime))
    AST = np.zeros((nx, ntime))
    TMP = np.zeros((nx, ntime))

    if double_ended_flag:
        REV_ST = np.zeros((nx, ntime))
        REV_AST = np.zeros((nx, ntime))

    for ii in range(ntime):
        data, meta = read_sensornet_single(filepathlist[ii])

        timestamp[ii] = pd.DatetimeIndex([meta["date"] + " " + meta["time"]])[0]
        probe1temperature[ii] = float(meta["T ext. ref 1 (°C)"])
        probe2temperature[ii] = float(meta["T ext. ref 2 (°C)"])
        referenceTemperature[ii] = float(meta["T internal ref (°C)"])
        gamma_ddf[ii] = float(meta["gamma"])
        k_internal[ii] = float(meta["k internal"])
        k_external[ii] = float(meta["k external"])
        acquisitiontimeFW[ii] = float(meta["forward acquisition time"])
        acquisitiontimeBW[ii] = float(meta["reverse acquisition time"])

        ST[:, ii] = data["st"]
        AST[:, ii] = data["ast"]
        TMP[:, ii] = data["tmp"]

        if double_ended_flag:
            REV_ST[:, ii] = data["rst"]
            REV_AST[:, ii] = data["rast"]

    if fiber_length is None and double_ended_flag:
        fiber_length = np.max([0.0, xraw[-1] - add_internal_fiber_length])
    elif fiber_length is None and not double_ended_flag:
        fiber_length = xraw[-1]
    else:
        pass

    assert fiber_length > 0.0, (
        "`fiber_length` is not defined. Use key"
        "word argument in read function." + str(fiber_length)
    )

    fiber_start_index = (np.abs(xraw + add_internal_fiber_length)).argmin()
    fiber_0_index = np.abs(xraw).argmin()
    fiber_1_index = (np.abs(xraw - fiber_length)).argmin()
    fiber_n_indices = fiber_1_index - fiber_0_index
    fiber_n_indices_internal = fiber_0_index - fiber_start_index
    if double_ended_flag:
        fiber_end_index = np.min([xraw.size, fiber_1_index + fiber_n_indices_internal])
    else:
        fiber_end_index = fiber_1_index

    if double_ended_flag:
        if not flip_reverse_measurements:
            # fiber length how the backward channel is aligned
            fiber_length_raw = float(meta["fibre end"])
            fiber_bw_1_index = np.abs(xraw - fiber_length_raw).argmin()
            fiber_bw_end_index = np.min(
                [xraw.size, fiber_bw_1_index + (fiber_end_index - fiber_1_index)]
            )
            fiber_bw_start_index = np.max(
                [0, fiber_bw_1_index - fiber_n_indices - fiber_n_indices_internal]
            )

            if (fiber_end_index - fiber_start_index) == (
                fiber_bw_end_index - fiber_bw_start_index
            ):
                REV_ST = REV_ST[fiber_bw_start_index:fiber_bw_end_index]
                REV_AST = REV_AST[fiber_bw_start_index:fiber_bw_end_index]
            else:
                REV_ST = REV_ST[fiber_start_index:fiber_end_index]
                REV_AST = REV_AST[fiber_start_index:fiber_end_index]

        else:
            # Use the fiber indices from the forward channel
            n_indices_internal_left = fiber_0_index - fiber_start_index
            n_indices_internal_right = np.max([0, fiber_end_index - fiber_1_index])
            n_indices_internal_shortest = np.min(
                [n_indices_internal_left, n_indices_internal_right]
            )
            fiber_start_index = fiber_0_index - n_indices_internal_shortest
            fiber_end_index = (
                fiber_0_index + fiber_n_indices + n_indices_internal_shortest
            )
            REV_ST = REV_ST[fiber_end_index:fiber_start_index:-1]
            REV_AST = REV_AST[fiber_end_index:fiber_start_index:-1]

    x = xraw[fiber_start_index:fiber_end_index]
    TMP = TMP[fiber_start_index:fiber_end_index]
    ST = ST[fiber_start_index:fiber_end_index]
    AST = AST[fiber_start_index:fiber_end_index]

    data_vars = {
        "st": (["x", "time"], ST, dim_attrs["st"]),
        "ast": (["x", "time"], AST, dim_attrs["ast"]),
        "tmp": (["x", "time"], TMP, dim_attrs["tmp"]),
        "probe1Temperature": (
            "time",
            probe1temperature,
            {
                "name": "Probe 1 temperature",
                "description": "reference probe 1 " "temperature",
                "units": r"$^\circ$C",
            },
        ),
        "probe2Temperature": (
            "time",
            probe2temperature,
            {
                "name": "Probe 2 temperature",
                "description": "reference probe 2 " "temperature",
                "units": r"$^\circ$C",
            },
        ),
        "referenceTemperature": (
            "time",
            referenceTemperature,
            {
                "name": "reference temperature",
                "description": "Internal reference " "temperature",
                "units": r"$^\circ$C",
            },
        ),
        "gamma_ddf": (
            "time",
            gamma_ddf,
            {
                "name": "gamma ddf",
                "description": "machine " "calibrated gamma",
                "units": "-",
            },
        ),
        "k_internal": (
            "time",
            k_internal,
            {
                "name": "k internal",
                "description": "machine calibrated " "internal k",
                "units": "-",
            },
        ),
        "k_external": (
            "time",
            k_external,
            {
                "name": "reference temperature",
                "description": "machine calibrated " "external k",
                "units": "-",
            },
        ),
        "userAcquisitionTimeFW": (
            "time",
            acquisitiontimeFW,
            dim_attrs["userAcquisitionTimeFW"],
        ),
        "userAcquisitionTimeBW": (
            "time",
            acquisitiontimeBW,
            dim_attrs["userAcquisitionTimeBW"],
        ),
    }

    if double_ended_flag:
        data_vars["rst"] = (["x", "time"], REV_ST, dim_attrs["rst"])
        data_vars["rast"] = (["x", "time"], REV_AST, dim_attrs["rast"])

    filenamelist = [os.path.split(f)[-1] for f in filepathlist]

    coords = {"x": ("x", x, dim_attrs["x"]), "filename": ("time", filenamelist)}

    dtFW = data_vars["userAcquisitionTimeFW"][1].astype("timedelta64[s]")
    dtBW = data_vars["userAcquisitionTimeBW"][1].astype("timedelta64[s]")
    if not double_ended_flag:
        tcoords = coords_time(
            np.array(timestamp).astype("datetime64[ns]"),
            timezone_netcdf=timezone_netcdf,
            timezone_input_files=timezone_input_files,
            dtFW=dtFW,
            double_ended_flag=double_ended_flag,
        )
    else:
        tcoords = coords_time(
            np.array(timestamp).astype("datetime64[ns]"),
            timezone_netcdf=timezone_netcdf,
            timezone_input_files=timezone_input_files,
            dtFW=dtFW,
            dtBW=dtBW,
            double_ended_flag=double_ended_flag,
        )

    coords.update(tcoords)

    return data_vars, coords, attrs



import os
from collections import namedtuple
from pathlib import Path
from typing import Optional

import nbformat
import pytest
from nbconvert.preprocessors import CellExecutionError
from nbconvert.preprocessors import ExecutePreprocessor

NBError = namedtuple("NBError", "title, errname, errvalue, exception")

wd = os.path.dirname(os.path.abspath(__file__))
src_dir = os.path.join(wd, "..", "docs", "notebooks")


@pytest.mark.parametrize(
    "src_path", Path(src_dir).glob("*.ipynb"), ids=lambda x: x.name
)
def test_docs_notebook(src_path):
    print(_test_notebook(src_path, "python3"))


@pytest.mark.xfail
def test_identify_not_working_docs_notebook():
    fp = Path(os.path.join(wd, "data", "docs_notebooks", "01Not_working.ipynb"))
    _test_notebook(fp, "python3")


def _test_notebook(notebook_file, kernel) -> Optional[NBError]:
    """
    Test single notebook.

    Parameters
    ----------
    notebook_file : str
        Source path for notebook
    kernel : str, optional
        Notebook kernel name, by default "python3"

    Returns
    -------
    NBError
        Error tuple

    """
    print(f"\nTesting notebook {notebook_file.name}")
    with open(notebook_file, "rb") as file_handle:
        nb_bytes = file_handle.read()
    nb_text = nb_bytes.decode("utf-8")
    nb_content = nbformat.reads(nb_text, as_version=4)
    exec_processor = ExecutePreprocessor(timeout=600, kernel_name=kernel)

    try:
        print(f"{notebook_file.name} - running notebook: ...")
        exec_processor.preprocess(
            nb_content, {"metadata": {"path": str(notebook_file.parent)}}
        )

    except CellExecutionError as cell_err:
        msg = f"Error executing the notebook '{notebook_file.absolute()}"
        print(msg)
        return NBError(
            f"Error while running notebook {notebook_file}",
            cell_err.ename,
            cell_err.evalue,
            cell_err.args[0],
        )

    return None



import warnings

import numpy as np
import scipy.sparse as sp
from scipy.optimize import minimize
from scipy.sparse import linalg as ln


def variance_stokes_constant_helper(data_dict):
    def func_fit(p, xs):
        return p[:xs, None] * p[None, xs:]

    def func_cost(p, data, xs):
        fit = func_fit(p, xs)
        return np.sum((fit - data) ** 2)

    resid_list = []

    for k, v in data_dict.items():
        for vi in v:
            nxs, nt = vi.shape
            npar = nt + nxs

            p1 = np.ones(npar) * vi.mean() ** 0.5

            res = minimize(func_cost, p1, args=(vi, nxs), method="Powell")
            assert res.success, "Unable to fit. Try variance_stokes_exponential"

            fit = func_fit(res.x, nxs)
            resid_list.append(fit - vi)

    resid = np.concatenate(resid_list)

    # unbiased estimater ddof=1, originally thought it was npar
    var_I = resid.var(ddof=1)

    return var_I, resid


def variance_stokes_exponential_helper(
    nt, x, y, len_stretch_list, use_statsmodels, suppress_info
):
    n_sections = len(len_stretch_list)  # number of sections
    n_locs = sum(len_stretch_list)  # total number of locations along cable used
    # for reference.

    data1 = x
    data2 = np.ones(sum(len_stretch_list) * nt)
    data = np.concatenate([data1, data2])

    # alpha is NOT the same for all -> one column per section
    coords1row = np.arange(nt * n_locs)
    coords1col = np.hstack(
        [np.ones(in_locs * nt) * i for i, in_locs in enumerate(len_stretch_list)]
    )  # C for

    # second calibration parameter is different per section and per timestep
    coords2row = np.arange(nt * n_locs)
    coords2col = np.hstack(
        [
            np.repeat(
                np.arange(i * nt + n_sections, (i + 1) * nt + n_sections), in_locs
            )
            for i, in_locs in enumerate(len_stretch_list)
        ]
    )  # C for
    coords = (
        np.concatenate([coords1row, coords2row]),
        np.concatenate([coords1col, coords2col]),
    )

    lny = np.log(y)
    w = y.copy()  # 1/std.

    ddof = n_sections + nt * n_sections  # see numpy documentation on ddof

    if use_statsmodels:
        # returns the same answer with statsmodel
        import statsmodels.api as sm

        X = sp.coo_matrix(
            (data, coords), shape=(nt * n_locs, ddof), dtype=float, copy=False
        )

        mod_wls = sm.WLS(lny, X.toarray(), weights=w**2)
        res_wls = mod_wls.fit()
        # print(res_wls.summary())
        a = res_wls.params

    else:
        wdata = data * np.hstack((w, w))
        wX = sp.coo_matrix(
            (wdata, coords),
            shape=(nt * n_locs, n_sections + nt * n_sections),
            dtype=float,
            copy=False,
        )

        wlny = lny * w

        p0_est = np.asarray(n_sections * [0.0] + nt * n_sections * [8])
        # noinspection PyTypeChecker
        a = ln.lsqr(wX, wlny, x0=p0_est, show=not suppress_info, calc_var=False)[0]

    beta = a[:n_sections]
    beta_expand_to_sec = np.hstack(
        [
            np.repeat(float(beta[i]), leni * nt)
            for i, leni in enumerate(len_stretch_list)
        ]
    )
    G = np.asarray(a[n_sections:])
    G_expand_to_sec = np.hstack(
        [
            np.repeat(G[i * nt : (i + 1) * nt], leni)
            for i, leni in enumerate(len_stretch_list)
        ]
    )

    I_est = np.exp(G_expand_to_sec) * np.exp(x * beta_expand_to_sec)
    resid = I_est - y
    var_I = resid.var(ddof=1)
    return var_I, resid


def variance_stokes_linear_helper(st_sec, resid_sec, nbin, through_zero):
    # Adjust nbin silently to fit residuals in
    # rectangular matrix and use numpy for computation
    nbin_ = nbin
    while st_sec.size % nbin_:
        nbin_ -= 1

    if nbin_ != nbin:
        print(f"Adjusting nbin to: {nbin_} to fit residuals in ")
        nbin = nbin_

    isort = np.argsort(st_sec)
    st_sort_mean = st_sec[isort].reshape((nbin, -1)).mean(axis=1)
    st_sort_var = resid_sec[isort].reshape((nbin, -1)).var(axis=1)

    if through_zero:
        # VAR(Stokes) = slope * Stokes
        offset = 0.0
        slope = np.linalg.lstsq(st_sort_mean[:, None], st_sort_var, rcond=None)[0]

    else:
        # VAR(Stokes) = slope * Stokes + offset
        slope, offset = np.linalg.lstsq(
            np.hstack((st_sort_mean[:, None], np.ones((nbin, 1)))),
            st_sort_var,
            rcond=None,
        )[0]

        if offset < 0:
            warnings.warn(
                "Warning! Offset of variance_stokes_linear() "
                "is negative. This is phisically "
                "not possible. Most likely, your Stokes intensities do "
                "not vary enough to fit a linear curve. Either "
                "use `through_zero` option or use "
                "`variance_stokes_constant()`. Another reason "
                "could be that your sections are defined to be "
                "wider than they actually are."
            )

    def var_fun(stokes):
        return slope * stokes + offset

    return slope, offset, st_sort_mean, st_sort_var, resid_sec, var_fun


def check_allclose_acquisitiontime(acquisitiontime, eps: float = 0.05) -> None:
    """Check if all acquisition times are of equal duration. For now it is not possible to calibrate
    over timesteps if the acquisition time of timesteps varies, as the Stokes variance
    would change over time.

    The acquisition time is stored for single ended measurements in userAcquisitionTime,
    for double ended measurements in userAcquisitionTimeFW and userAcquisitionTimeBW.

    Parameters
    ----------
    ds : DataStore
    eps : float
        Default accepts 1% of relative variation between min and max acquisition time.

    Returns:
    --------
    """
    dtmin = acquisitiontime.min()
    dtmax = acquisitiontime.max()
    dtavg = (dtmin + dtmax) / 2
    assert (
        dtmax - dtmin
    ) / dtavg < eps, "Acquisition time is Forward channel not equal for all time steps"
    pass



import os
import warnings

import numpy as np
import pytest
import scipy.sparse as sp
from scipy import stats
from xarray import Dataset

from dtscalibration.calibrate_utils import wls_sparse
from dtscalibration.calibrate_utils import wls_stats
from dtscalibration.variance_stokes import variance_stokes_exponential

np.random.seed(0)

fn = [
    "channel 1_20170921112245510.xml",
    "channel 1_20170921112746818.xml",
    "channel 1_20170921112746818.xml",
]
fn_single = [
    "channel 2_20180504132202074.xml",
    "channel 2_20180504132232903.xml",
    "channel 2_20180504132303723.xml",
]

if 1:
    # working dir is tests
    wd = os.path.dirname(os.path.abspath(__file__))
    data_dir_single_ended = os.path.join(wd, "data", "single_ended")
    data_dir_double_ended = os.path.join(wd, "data", "double_ended")
    data_dir_double_ended2 = os.path.join(wd, "data", "double_ended2")

else:
    # working dir is src
    data_dir_single_ended = os.path.join("..", "..", "tests", "data", "single_ended")
    data_dir_double_ended = os.path.join("..", "..", "tests", "data", "double_ended")
    data_dir_double_ended2 = os.path.join("..", "..", "tests", "data", "double_ended2")


def assert_almost_equal_verbose(actual, desired, verbose=False, **kwargs):
    """Print the actual precision decimals"""
    err = np.abs(actual - desired).max()

    with warnings.catch_warnings():
        # Supress zero division warnings
        warnings.filterwarnings("ignore", message="divide by zero encountered in log10")
        dec = -np.ceil(np.log10(err))

    if not (np.isfinite(dec)):
        dec = 18.0

    m = "\n>>>>>The actual precision is: " + str(float(dec))

    if verbose:
        print(m)

    desired2 = np.broadcast_to(desired, actual.shape)
    np.testing.assert_almost_equal(actual, desired2, err_msg=m, **kwargs)
    pass


def test_double_ended_wls_estimate_synthetic():
    """Checks whether the coefficients are correctly defined by creating a
    synthetic measurement set, and derive the parameters from this set.
    Without variance.
    They should be the same as the parameters used to create the synthetic
    measurment set"""
    cable_len = 100.0
    nt = 50
    time = np.arange(nt)
    x = np.linspace(0.0, cable_len, 100)
    ts_cold = np.ones(nt) * 4.0
    ts_warm = np.ones(nt) * 20.0

    C_p = 15246
    C_m = 2400.0
    dalpha_r = 0.0005284
    dalpha_m = 0.0004961
    dalpha_p = 0.0005607
    gamma = 482.6
    cold_mask = x < 0.5 * cable_len
    warm_mask = np.invert(cold_mask)  # == False
    temp_real = np.ones((len(x), nt))
    temp_real[cold_mask] *= ts_cold + 273.15
    temp_real[warm_mask] *= ts_warm + 273.15

    st = (
        C_p
        * np.exp(-(dalpha_r + dalpha_p) * x[:, None])
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    ast = (
        C_m
        * np.exp(-(dalpha_r + dalpha_m) * x[:, None])
        / (np.exp(gamma / temp_real) - 1)
    )
    rst = (
        C_p
        * np.exp(-(dalpha_r + dalpha_p) * (-x[:, None] + cable_len))
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    rast = (
        C_m
        * np.exp(-(dalpha_r + dalpha_m) * (-x[:, None] + cable_len))
        / (np.exp(gamma / temp_real) - 1)
    )

    alpha = np.mean(np.log(rst / rast) - np.log(st / ast), axis=1) / 2
    alpha -= alpha[0]  # the first x-index is where to start counting

    ds = Dataset(
        {
            "st": (["x", "time"], st),
            "ast": (["x", "time"], ast),
            "rst": (["x", "time"], rst),
            "rast": (["x", "time"], rast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "userAcquisitionTimeBW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "1"},
    )

    sections = {
        "cold": [slice(0.0, 0.4 * cable_len)],
        "warm": [slice(0.65 * cable_len, cable_len)],
    }

    # WLS
    out = ds.dts.calibrate_double_ended(
        sections=sections,
        st_var=1e-7,
        ast_var=1e-7,
        rst_var=1e-7,
        rast_var=1e-7,
        method="wls",
        solver="sparse",
    )

    assert_almost_equal_verbose(out["gamma"].values, gamma, decimal=10)
    assert_almost_equal_verbose(out["alpha"].values, alpha, decimal=8)
    assert_almost_equal_verbose(out["tmpf"].values, temp_real - 273.15, decimal=6)
    assert_almost_equal_verbose(out["tmpb"].values, temp_real - 273.15, decimal=6)
    assert_almost_equal_verbose(out["tmpw"].values, temp_real - 273.15, decimal=6)


def test_double_ended_wls_estimate_synthetic_df_and_db_are_different():
    """Checks whether the coefficients are correctly defined by creating a
    synthetic measurement set, and derive the parameters from this set.
    Without variance.
    They should be the same as the parameters used to create the synthetic
    measurment set. This one has a different D for the forward channel than
    for the backward channel."""

    cable_len = 100.0
    nt = 3
    time = np.arange(nt)
    x = np.linspace(0.0, cable_len, 8)
    ts_cold = np.ones(nt) * 4.0 + np.cos(time) * 4
    ts_warm = np.ones(nt) * 20.0 + -np.sin(time) * 4

    C_p = 1324  # 1/2 * E0 * v * K_+/lam_+^4
    eta_pf = np.cos(time) / 10 + 1  # eta_+ (gain factor forward channel)
    eta_pb = np.sin(time) / 10 + 1  # eta_- (gain factor backward channel)
    C_m = 5000.0
    eta_mf = np.cos(time + np.pi / 8) / 10 + 1
    eta_mb = np.sin(time + np.pi / 8) / 10 + 1
    dalpha_r = 0.005284
    dalpha_m = 0.004961
    dalpha_p = 0.005607
    gamma = 482.6

    temp_real_kelvin = np.zeros((len(x), nt)) + 273.15
    temp_real_kelvin[x < 0.2 * cable_len] += ts_cold[None]
    temp_real_kelvin[x > 0.85 * cable_len] += ts_warm[None]
    temp_real_celsius = temp_real_kelvin - 273.15

    st = (
        eta_pf[None]
        * C_p
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_p * x[:, None])
        * np.exp(gamma / temp_real_kelvin)
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    ast = (
        eta_mf[None]
        * C_m
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_m * x[:, None])
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    rst = (
        eta_pb[None]
        * C_p
        * np.exp(-dalpha_r * (-x[:, None] + cable_len))
        * np.exp(-dalpha_p * (-x[:, None] + cable_len))
        * np.exp(gamma / temp_real_kelvin)
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    rast = (
        eta_mb[None]
        * C_m
        * np.exp(-dalpha_r * (-x[:, None] + cable_len))
        * np.exp(-dalpha_m * (-x[:, None] + cable_len))
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )

    c_f = np.log(eta_mf * C_m / (eta_pf * C_p))
    c_b = np.log(eta_mb * C_m / (eta_pb * C_p))

    dalpha = dalpha_p - dalpha_m  # \Delta\alpha
    alpha_int = cable_len * dalpha

    df = c_f  # reference section starts at first x-index
    db = c_b + alpha_int
    i_fw = np.log(st / ast)
    i_bw = np.log(rst / rast)

    E_real = (i_bw - i_fw) / 2 + (db - df) / 2

    ds = Dataset(
        {
            "st": (["x", "time"], st),
            "ast": (["x", "time"], ast),
            "rst": (["x", "time"], rst),
            "rast": (["x", "time"], rast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "userAcquisitionTimeBW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "1"},
    )

    sections = {
        "cold": [slice(0.0, 0.09 * cable_len)],
        "warm": [slice(0.9 * cable_len, cable_len)],
    }

    real_ans2 = np.concatenate(([gamma], df, db, E_real[:, 0]))

    out = ds.dts.calibrate_double_ended(
        sections=sections,
        st_var=1.5,
        ast_var=1.5,
        rst_var=1.0,
        rast_var=1.0,
        method="wls",
        solver="sparse",
        fix_gamma=(gamma, 0.0),
    )

    assert_almost_equal_verbose(df, out["df"].values, decimal=14)
    assert_almost_equal_verbose(db, out["db"].values, decimal=13)
    assert_almost_equal_verbose(
        x * (dalpha_p - dalpha_m),
        out["alpha"].values - out["alpha"].values[0],
        decimal=13,
    )
    assert np.all(np.abs(real_ans2 - out["p_val"].values) < 1e-10)
    assert_almost_equal_verbose(temp_real_celsius, out["tmpf"].values, decimal=10)
    assert_almost_equal_verbose(temp_real_celsius, out["tmpb"].values, decimal=10)
    assert_almost_equal_verbose(temp_real_celsius, out["tmpw"].values, decimal=10)
    pass


@pytest.mark.xfail
def test_fail_if_st_labels_are_passed_to_calibration_function():
    """Same as
    `test_double_ended_wls_estimate_synthetic_df_and_db_are_different`
    Which runs fast."""

    cable_len = 100.0
    nt = 3
    time = np.arange(nt)
    x = np.linspace(0.0, cable_len, 8)
    ts_cold = np.ones(nt) * 4.0 + np.cos(time) * 4
    ts_warm = np.ones(nt) * 20.0 + -np.sin(time) * 4

    C_p = 1324  # 1/2 * E0 * v * K_+/lam_+^4
    eta_pf = np.cos(time) / 10 + 1  # eta_+ (gain factor forward channel)
    eta_pb = np.sin(time) / 10 + 1  # eta_- (gain factor backward channel)
    C_m = 5000.0
    eta_mf = np.cos(time + np.pi / 8) / 10 + 1
    eta_mb = np.sin(time + np.pi / 8) / 10 + 1
    dalpha_r = 0.005284
    dalpha_m = 0.004961
    dalpha_p = 0.005607
    gamma = 482.6

    temp_real_kelvin = np.zeros((len(x), nt)) + 273.15
    temp_real_kelvin[x < 0.2 * cable_len] += ts_cold[None]
    temp_real_kelvin[x > 0.85 * cable_len] += ts_warm[None]

    st = (
        eta_pf[None]
        * C_p
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_p * x[:, None])
        * np.exp(gamma / temp_real_kelvin)
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    ast = (
        eta_mf[None]
        * C_m
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_m * x[:, None])
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    rst = (
        eta_pb[None]
        * C_p
        * np.exp(-dalpha_r * (-x[:, None] + cable_len))
        * np.exp(-dalpha_p * (-x[:, None] + cable_len))
        * np.exp(gamma / temp_real_kelvin)
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    rast = (
        eta_mb[None]
        * C_m
        * np.exp(-dalpha_r * (-x[:, None] + cable_len))
        * np.exp(-dalpha_m * (-x[:, None] + cable_len))
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )

    ds = Dataset(
        {
            "ST": (["x", "time"], st),
            "AST": (["x", "time"], ast),
            "REV-ST": (["x", "time"], rst),
            "REV-AST": (["x", "time"], rast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "userAcquisitionTimeBW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "1"},
    )
    ds = ds.rename_labels()

    sections = {
        "cold": [slice(0.0, 0.09 * cable_len)],
        "warm": [slice(0.9 * cable_len, cable_len)],
    }

    ds.dts.calibrate_double_ended(
        sections=sections,
        st_label="ST",
        ast_label="AST",
        rst_label="REV-ST",
        rast_label="REV-AST",
        st_var=1.5,
        ast_var=1.5,
        rst_var=1.0,
        rast_var=1.0,
        method="wls",
        solver="sparse",
        mc_sample_size=1000,
        fix_gamma=(gamma, 0.0),
        mc_remove_set_flag=True,
    )
    pass


def test_double_ended_asymmetrical_attenuation():
    cable_len = 100.0
    nt = 3
    time = np.arange(nt)
    nx_per_sec = 1
    nx = nx_per_sec * 8
    x = np.linspace(0.0, cable_len, nx)
    ts_cold = 4.0 + np.cos(time) * 4
    ts_warm = 20.0 + -np.sin(time) * 4
    ts_ground = 6.0

    C_p = 1324  # 1/2 * E0 * v * K_+/lam_+^4
    eta_pf = np.cos(time) / 10 + 1  # eta_+ (gain factor forward channel)
    eta_pb = np.sin(time) / 10 + 1  # eta_- (gain factor backward channel)
    C_m = 5000.0
    eta_mf = np.cos(time + np.pi / 8) / 10 + 1
    eta_mb = np.sin(time + np.pi / 8) / 10 + 1
    dalpha_r = 0.005284
    dalpha_m = 0.004961
    dalpha_p = 0.005607
    gamma = 482.6
    talph_fw = 0.9
    talph_bw = 0.8

    temp_real_kelvin = np.zeros((len(x), nt)) + 273.15
    temp_real_kelvin[:nx_per_sec] += ts_cold[None]
    temp_real_kelvin[nx_per_sec : 2 * nx_per_sec] += ts_warm[None]
    temp_real_kelvin[-nx_per_sec:] += ts_cold[None]
    temp_real_kelvin[-2 * nx_per_sec : -nx_per_sec] += ts_warm[None]
    temp_real_kelvin[2 * nx_per_sec : -2 * nx_per_sec] += ts_ground
    temp_real_celsius = temp_real_kelvin - 273.15

    st = (
        eta_pf[None]
        * C_p
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_p * x[:, None])
        * np.exp(gamma / temp_real_kelvin)
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    st[4 * nx_per_sec :] *= talph_fw
    ast = (
        eta_mf[None]
        * C_m
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_m * x[:, None])
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    rst = (
        eta_pb[None]
        * C_p
        * np.exp(-dalpha_r * (-x[:, None] + cable_len))
        * np.exp(-dalpha_p * (-x[:, None] + cable_len))
        * np.exp(gamma / temp_real_kelvin)
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    rst[: 4 * nx_per_sec] *= talph_bw
    rast = (
        eta_mb[None]
        * C_m
        * np.exp(-dalpha_r * (-x[:, None] + cable_len))
        * np.exp(-dalpha_m * (-x[:, None] + cable_len))
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )

    ds = Dataset(
        {
            "TMPR": (["x", "time"], temp_real_celsius),
            "st": (["x", "time"], st),
            "ast": (["x", "time"], ast),
            "rst": (["x", "time"], rst),
            "rast": (["x", "time"], rast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "userAcquisitionTimeBW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
    )

    sections = {
        "cold": [slice(0.0, x[nx_per_sec - 1]), slice(x[-nx_per_sec], x[-1])],
        "warm": [
            slice(x[nx_per_sec], x[2 * nx_per_sec - 1]),
            slice(x[-2 * nx_per_sec], x[-1 * nx_per_sec - 1]),
        ],
    }

    out = ds.dts.calibrate_double_ended(
        sections=sections,
        st_var=1.5,
        ast_var=1.5,
        rst_var=1.0,
        rast_var=1.0,
        method="wls",
        solver="sparse",
        trans_att=[50.0],
    )

    assert_almost_equal_verbose(temp_real_celsius, out.tmpf.values, decimal=7)
    assert_almost_equal_verbose(temp_real_celsius, out.tmpb.values, decimal=7)
    assert_almost_equal_verbose(temp_real_celsius, out.tmpw.values, decimal=7)
    pass


def test_double_ended_one_matching_section_and_one_asym_att():
    cable_len = 100.0
    nt = 3
    time = np.arange(nt)
    nx_per_sec = 2
    nx = nx_per_sec * 8
    x = np.linspace(0.0, cable_len, nx)
    ts_cold = 4.0 + np.cos(time) * 4
    ts_warm = 20.0 + -np.sin(time) * 4
    ts_ground = 6.0

    C_p = 1324  # 1/2 * E0 * v * K_+/lam_+^4
    eta_pf = np.cos(time) / 10 + 1  # eta_+ (gain factor forward channel)
    eta_pb = np.sin(time) / 10 + 1  # eta_- (gain factor backward channel)
    C_m = 5000.0
    eta_mf = np.cos(time + np.pi / 8) / 10 + 1
    eta_mb = np.sin(time + np.pi / 8) / 10 + 1
    dalpha_r = 0.005284
    dalpha_m = 0.004961
    dalpha_p = 0.005607
    gamma = 482.6
    talph_fw = 0.9
    talph_bw = 0.8

    temp_real_kelvin = np.zeros((len(x), nt)) + 273.15
    temp_real_kelvin[:nx_per_sec] += ts_cold[None]
    temp_real_kelvin[nx_per_sec : 2 * nx_per_sec] += ts_warm[None]
    temp_real_kelvin[-nx_per_sec:] += ts_cold[None]
    temp_real_kelvin[-2 * nx_per_sec : -nx_per_sec] += ts_warm[None]
    temp_real_kelvin[2 * nx_per_sec : -2 * nx_per_sec] += ts_ground
    temp_real_celsius = temp_real_kelvin - 273.15

    st = (
        eta_pf[None]
        * C_p
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_p * x[:, None])
        * np.exp(gamma / temp_real_kelvin)
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    st[4 * nx_per_sec :] *= talph_fw
    ast = (
        eta_mf[None]
        * C_m
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_m * x[:, None])
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    rst = (
        eta_pb[None]
        * C_p
        * np.exp(-dalpha_r * (-x[:, None] + cable_len))
        * np.exp(-dalpha_p * (-x[:, None] + cable_len))
        * np.exp(gamma / temp_real_kelvin)
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    rst[: 4 * nx_per_sec] *= talph_bw
    rast = (
        eta_mb[None]
        * C_m
        * np.exp(-dalpha_r * (-x[:, None] + cable_len))
        * np.exp(-dalpha_m * (-x[:, None] + cable_len))
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )

    ds = Dataset(
        {
            "TMPR": (["x", "time"], temp_real_celsius),
            "st": (["x", "time"], st),
            "ast": (["x", "time"], ast),
            "rst": (["x", "time"], rst),
            "rast": (["x", "time"], rast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "userAcquisitionTimeBW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "1"},
    )

    sections = {
        "cold": [slice(0.0, x[nx_per_sec - 1])],
        "warm": [slice(x[nx_per_sec], x[2 * nx_per_sec - 1])],
    }

    out = ds.dts.calibrate_double_ended(
        sections=sections,
        st_var=1.5,
        ast_var=1.5,
        rst_var=1.0,
        rast_var=1.0,
        method="wls",
        solver="sparse",
        trans_att=[50.0],
        matching_sections=[
            (
                slice(x[3 * nx_per_sec], x[4 * nx_per_sec - 1]),
                slice(x[4 * nx_per_sec], x[5 * nx_per_sec - 1]),
                True,
            )
        ],
    )

    assert_almost_equal_verbose(temp_real_celsius, out["tmpf"].values, decimal=7)
    assert_almost_equal_verbose(temp_real_celsius, out["tmpb"].values, decimal=7)
    assert_almost_equal_verbose(temp_real_celsius, out["tmpw"].values, decimal=7)


def test_double_ended_two_matching_sections_and_two_asym_atts():
    """Setup contains two matching sections and two connectors that introduce
    asymmetrical attenuation. Solves beautifully."""

    cable_len = 100.0
    nt = 5
    time = np.arange(nt)
    nx_per_sec = 4
    nx = nx_per_sec * 9
    x = np.linspace(0.0, cable_len, nx)
    ts_cold = 4.0 + np.cos(time) * 4
    ts_warm = 20.0 + -np.sin(time) * 4
    ts_ground = np.linspace(1, 9, num=nx_per_sec)

    C_p = 1324  # 1/2 * E0 * v * K_+/lam_+^4
    eta_pf = np.cos(time) / 10 + 1  # eta_+ (gain factor forward channel)
    eta_pb = np.sin(time) / 10 + 1  # eta_- (gain factor backward channel)
    C_m = 5000.0
    eta_mf = np.cos(time + np.pi / 8) / 10 + 1
    eta_mb = np.sin(time + np.pi / 8) / 10 + 1
    dalpha_r = 0.005284
    dalpha_m = 0.004961
    dalpha_p = 0.005607
    gamma = 482.6
    talph_fw = 0.95
    talph_bw = 0.85

    temp_real_kelvin = np.zeros((len(x), nt)) + 273.15
    temp_real_kelvin[:nx_per_sec] += ts_cold[None]
    temp_real_kelvin[nx_per_sec : 2 * nx_per_sec] += ts_warm[None]
    temp_real_kelvin[2 * nx_per_sec : 3 * nx_per_sec] += ts_ground[:, None]
    temp_real_kelvin[3 * nx_per_sec : 4 * nx_per_sec] += ts_ground[::-1, None]
    temp_real_kelvin[5 * nx_per_sec : 6 * nx_per_sec] += ts_ground[:, None] + 5
    temp_real_kelvin[6 * nx_per_sec : 7 * nx_per_sec] += ts_ground[:, None] + 5
    temp_real_kelvin[7 * nx_per_sec : 8 * nx_per_sec] += ts_warm[None]
    temp_real_kelvin[8 * nx_per_sec : 9 * nx_per_sec] += ts_cold[None]

    temp_real_celsius = temp_real_kelvin - 273.15

    st = (
        eta_pf[None]
        * C_p
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_p * x[:, None])
        * np.exp(gamma / temp_real_kelvin)
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    st[3 * nx_per_sec :] *= talph_fw
    st[6 * nx_per_sec :] *= talph_fw
    ast = (
        eta_mf[None]
        * C_m
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_m * x[:, None])
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    rst = (
        eta_pb[None]
        * C_p
        * np.exp(-dalpha_r * (-x[:, None] + cable_len))
        * np.exp(-dalpha_p * (-x[:, None] + cable_len))
        * np.exp(gamma / temp_real_kelvin)
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    rst[: 3 * nx_per_sec] *= talph_bw
    rst[: 6 * nx_per_sec] *= talph_bw
    rast = (
        eta_mb[None]
        * C_m
        * np.exp(-dalpha_r * (-x[:, None] + cable_len))
        * np.exp(-dalpha_m * (-x[:, None] + cable_len))
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )

    ds = Dataset(
        {
            "TMPR": (["x", "time"], temp_real_celsius),
            "st": (["x", "time"], st),
            "ast": (["x", "time"], ast),
            "rst": (["x", "time"], rst),
            "rast": (["x", "time"], rast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "userAcquisitionTimeBW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "1"},
    )

    sections = {
        "cold": [slice(0.0, x[nx_per_sec - 1])],
        "warm": [slice(x[nx_per_sec], x[2 * nx_per_sec - 1])],
    }
    ms = [
        (
            slice(x[2 * nx_per_sec], x[3 * nx_per_sec - 1]),
            slice(x[3 * nx_per_sec], x[4 * nx_per_sec - 1]),
            True,
        ),
        (
            slice(x[5 * nx_per_sec], x[6 * nx_per_sec - 1]),
            slice(x[6 * nx_per_sec], x[7 * nx_per_sec - 1]),
            False,
        ),
    ]

    out = ds.dts.calibrate_double_ended(
        sections=sections,
        st_var=0.5,
        ast_var=0.5,
        rst_var=0.1,
        rast_var=0.1,
        method="wls",
        solver="sparse",
        trans_att=[x[3 * nx_per_sec], x[6 * nx_per_sec]],
        matching_sections=ms,
    )

    assert_almost_equal_verbose(temp_real_celsius, out["tmpf"].values, decimal=7)
    assert_almost_equal_verbose(temp_real_celsius, out["tmpb"].values, decimal=7)
    assert_almost_equal_verbose(temp_real_celsius, out["tmpw"].values, decimal=7)
    pass


def test_double_ended_wls_fix_gamma_estimate_synthetic():
    """Checks whether the coefficients are correctly defined by creating a
    synthetic measurement set, and derive the parameters from this set.
    Without variance.
    They should be the same as the parameters used to create the synthetic
    measurment set"""

    cable_len = 100.0
    nt = 500
    time = np.arange(nt)
    x = np.linspace(0.0, cable_len, 100)
    ts_cold = np.ones(nt) * 4.0
    ts_warm = np.ones(nt) * 20.0

    C_p = 15246
    C_m = 2400.0
    dalpha_r = 0.0005284
    dalpha_m = 0.0004961
    dalpha_p = 0.0005607
    gamma = 482.6
    cold_mask = x < 0.5 * cable_len
    warm_mask = np.invert(cold_mask)  # == False
    temp_real = np.ones((len(x), nt))
    temp_real[cold_mask] *= ts_cold + 273.15
    temp_real[warm_mask] *= ts_warm + 273.15

    st = (
        C_p
        * np.exp(-(dalpha_r + dalpha_p) * x[:, None])
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    ast = (
        C_m
        * np.exp(-(dalpha_r + dalpha_m) * x[:, None])
        / (np.exp(gamma / temp_real) - 1)
    )
    rst = (
        C_p
        * np.exp(-(dalpha_r + dalpha_p) * (-x[:, None] + cable_len))
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    rast = (
        C_m
        * np.exp(-(dalpha_r + dalpha_m) * (-x[:, None] + cable_len))
        / (np.exp(gamma / temp_real) - 1)
    )

    alpha = np.mean(np.log(rst / rast) - np.log(st / ast), axis=1) / 2
    alpha -= alpha[0]  # the first x-index is where to start counting
    dalpha = dalpha_p - dalpha_m
    alpha2 = x * dalpha

    # to ensure the st, rst, ast, rast were correctly defined.
    np.testing.assert_allclose(alpha2, alpha, atol=1e-15, rtol=0)

    ds = Dataset(
        {
            "st": (["x", "time"], st),
            "ast": (["x", "time"], ast),
            "rst": (["x", "time"], rst),
            "rast": (["x", "time"], rast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "userAcquisitionTimeBW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "1"},
    )

    sections = {
        "cold": [slice(0.0, 0.35 * cable_len)],
        "warm": [slice(0.67 * cable_len, cable_len)],
    }

    # WLS
    out = ds.dts.calibrate_double_ended(
        sections=sections,
        st_var=1e-12,
        ast_var=1e-12,
        rst_var=1e-12,
        rast_var=1e-12,
        method="wls",
        solver="sparse",
        fix_gamma=(gamma, 0.0),
    )

    assert_almost_equal_verbose(out["gamma"].values, gamma, decimal=18)
    assert_almost_equal_verbose(out["alpha"].values, alpha, decimal=9)
    assert_almost_equal_verbose(out["tmpf"].values, temp_real - 273.15, decimal=6)
    assert_almost_equal_verbose(out["tmpb"].values, temp_real - 273.15, decimal=6)
    assert_almost_equal_verbose(out["tmpw"].values, temp_real - 273.15, decimal=6)

    pass


def test_double_ended_wls_fix_alpha_estimate_synthetic():
    """Checks whether the coefficients are correctly defined by creating a
    synthetic measurement set, and derive the parameters from this set.
    Without variance.
    They should be the same as the parameters used to create the synthetic
    measurment set"""

    cable_len = 100.0
    nt = 500
    time = np.arange(nt)
    x = np.linspace(0.0, cable_len, 100)
    ts_cold = np.ones(nt) * 4.0
    ts_warm = np.ones(nt) * 20.0

    C_p = 15246
    C_m = 2400.0
    dalpha_r = 0.0005284
    dalpha_m = 0.0004961
    dalpha_p = 0.0005607
    gamma = 482.6
    cold_mask = x < 0.5 * cable_len
    warm_mask = np.invert(cold_mask)  # == False
    temp_real = np.ones((len(x), nt))
    temp_real[cold_mask] *= ts_cold + 273.15
    temp_real[warm_mask] *= ts_warm + 273.15

    st = (
        C_p
        * np.exp(-(dalpha_r + dalpha_p) * x[:, None])
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    ast = (
        C_m
        * np.exp(-(dalpha_r + dalpha_m) * x[:, None])
        / (np.exp(gamma / temp_real) - 1)
    )
    rst = (
        C_p
        * np.exp(-(dalpha_r + dalpha_p) * (-x[:, None] + cable_len))
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    rast = (
        C_m
        * np.exp(-(dalpha_r + dalpha_m) * (-x[:, None] + cable_len))
        / (np.exp(gamma / temp_real) - 1)
    )

    alpha = np.mean(np.log(rst / rast) - np.log(st / ast), axis=1) / 2
    alpha -= alpha[0]  # the first x-index is where to start counting

    ds = Dataset(
        {
            "st": (["x", "time"], st),
            "ast": (["x", "time"], ast),
            "rst": (["x", "time"], rst),
            "rast": (["x", "time"], rast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "userAcquisitionTimeBW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "1"},
    )

    sections = {
        "cold": [slice(0.0, 0.4 * cable_len)],
        "warm": [slice(0.78 * cable_len, cable_len)],
    }

    # WLS
    out = ds.dts.calibrate_double_ended(
        sections=sections,
        st_var=1e-7,
        ast_var=1e-7,
        rst_var=1e-7,
        rast_var=1e-7,
        method="wls",
        solver="sparse",
        fix_alpha=(alpha, np.zeros_like(alpha)),
    )

    assert_almost_equal_verbose(out["gamma"].values, gamma, decimal=8)
    assert_almost_equal_verbose(out["alpha"].values, alpha, decimal=18)
    assert_almost_equal_verbose(out["tmpf"].values, temp_real - 273.15, decimal=7)
    assert_almost_equal_verbose(out["tmpb"].values, temp_real - 273.15, decimal=7)
    assert_almost_equal_verbose(out["tmpw"].values, temp_real - 273.15, decimal=7)

    pass


def test_double_ended_wls_fix_alpha_fix_gamma_estimate_synthetic():
    """Checks whether the coefficients are correctly defined by creating a
    synthetic measurement set, and derive the parameters from this set.
    Without variance.
    They should be the same as the parameters used to create the synthetic
    measurment set"""

    cable_len = 100.0
    nt = 500
    time = np.arange(nt)
    x = np.linspace(0.0, cable_len, 100)
    ts_cold = np.ones(nt) * 4.0
    ts_warm = np.ones(nt) * 20.0

    C_p = 15246
    C_m = 2400.0
    dalpha_r = 0.0005284
    dalpha_m = 0.0004961
    dalpha_p = 0.0005607
    gamma = 482.6
    cold_mask = x < 0.5 * cable_len
    warm_mask = np.invert(cold_mask)  # == False
    temp_real = np.ones((len(x), nt))
    temp_real[cold_mask] *= ts_cold + 273.15
    temp_real[warm_mask] *= ts_warm + 273.15

    st = (
        C_p
        * np.exp(-(dalpha_r + dalpha_p) * x[:, None])
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    ast = (
        C_m
        * np.exp(-(dalpha_r + dalpha_m) * x[:, None])
        / (np.exp(gamma / temp_real) - 1)
    )
    rst = (
        C_p
        * np.exp(-(dalpha_r + dalpha_p) * (-x[:, None] + cable_len))
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    rast = (
        C_m
        * np.exp(-(dalpha_r + dalpha_m) * (-x[:, None] + cable_len))
        / (np.exp(gamma / temp_real) - 1)
    )

    alpha = np.mean(np.log(rst / rast) - np.log(st / ast), axis=1) / 2
    alpha -= alpha[0]  # the first x-index is where to start counting

    ds = Dataset(
        {
            "st": (["x", "time"], st),
            "ast": (["x", "time"], ast),
            "rst": (["x", "time"], rst),
            "rast": (["x", "time"], rast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "userAcquisitionTimeBW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "1"},
    )

    sections = {
        "cold": [slice(0.0, 0.5 * cable_len)],
        "warm": [slice(0.5 * cable_len, cable_len)],
    }

    # WLS
    out = ds.dts.calibrate_double_ended(
        sections=sections,
        st_var=1e-7,
        ast_var=1e-7,
        rst_var=1e-7,
        rast_var=1e-7,
        method="wls",
        solver="sparse",
        fix_gamma=(gamma, 0.0),
        fix_alpha=(alpha, np.zeros_like(alpha)),
    )

    assert_almost_equal_verbose(out["gamma"].values, gamma, decimal=18)
    assert_almost_equal_verbose(out["alpha"].values, alpha, decimal=18)
    assert_almost_equal_verbose(out["tmpf"].values, temp_real - 273.15, decimal=11)
    assert_almost_equal_verbose(out["tmpb"].values, temp_real - 273.15, decimal=11)
    assert_almost_equal_verbose(out["tmpw"].values, temp_real - 273.15, decimal=11)

    pass


def test_double_ended_fix_alpha_matching_sections_and_one_asym_att():
    cable_len = 100.0
    nt = 3
    time = np.arange(nt)
    nx_per_sec = 2
    nx = nx_per_sec * 8
    x = np.linspace(0.0, cable_len, nx)
    ts_cold = 4.0 + np.cos(time) * 4
    ts_warm = 20.0 + -np.sin(time) * 4
    ts_ground = 6.0

    C_p = 1324  # 1/2 * E0 * v * K_+/lam_+^4
    eta_pf = np.cos(time) / 10 + 1  # eta_+ (gain factor forward channel)
    eta_pb = np.sin(time) / 10 + 1  # eta_- (gain factor backward channel)
    C_m = 5000.0
    eta_mf = np.cos(time + np.pi / 8) / 10 + 1
    eta_mb = np.sin(time + np.pi / 8) / 10 + 1
    dalpha_r = 0.005284
    dalpha_m = 0.004961
    dalpha_p = 0.005607
    gamma = 482.6
    talph_fw = 0.9
    talph_bw = 0.8

    temp_real_kelvin = np.zeros((len(x), nt)) + 273.15
    temp_real_kelvin[:nx_per_sec] += ts_cold[None]
    temp_real_kelvin[nx_per_sec : 2 * nx_per_sec] += ts_warm[None]
    temp_real_kelvin[-nx_per_sec:] += ts_cold[None]
    temp_real_kelvin[-2 * nx_per_sec : -nx_per_sec] += ts_warm[None]
    temp_real_kelvin[2 * nx_per_sec : -2 * nx_per_sec] += ts_ground
    temp_real_celsius = temp_real_kelvin - 273.15

    st = (
        eta_pf[None]
        * C_p
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_p * x[:, None])
        * np.exp(gamma / temp_real_kelvin)
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    st[4 * nx_per_sec :] *= talph_fw
    ast = (
        eta_mf[None]
        * C_m
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_m * x[:, None])
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    rst = (
        eta_pb[None]
        * C_p
        * np.exp(-dalpha_r * (-x[:, None] + cable_len))
        * np.exp(-dalpha_p * (-x[:, None] + cable_len))
        * np.exp(gamma / temp_real_kelvin)
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    rst[: 4 * nx_per_sec] *= talph_bw
    rast = (
        eta_mb[None]
        * C_m
        * np.exp(-dalpha_r * (-x[:, None] + cable_len))
        * np.exp(-dalpha_m * (-x[:, None] + cable_len))
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )

    ds = Dataset(
        {
            "TMPR": (["x", "time"], temp_real_celsius),
            "st": (["x", "time"], st),
            "ast": (["x", "time"], ast),
            "rst": (["x", "time"], rst),
            "rast": (["x", "time"], rast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "userAcquisitionTimeBW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "1"},
    )

    sections = {
        "cold": [slice(0.0, x[nx_per_sec - 1])],
        "warm": [slice(x[nx_per_sec], x[2 * nx_per_sec - 1])],
    }

    out = ds.dts.calibrate_double_ended(
        sections=sections,
        st_var=1.5,
        ast_var=1.5,
        rst_var=1.0,
        rast_var=1.0,
        method="wls",
        solver="sparse",
        trans_att=[50.0],
        matching_sections=[
            (
                slice(x[3 * nx_per_sec], x[4 * nx_per_sec - 1]),
                slice(x[4 * nx_per_sec], x[5 * nx_per_sec - 1]),
                True,
            )
        ],
    )

    # remove TA vars
    k = ["talpha_fw", "talpha_bw", "trans_att"]

    for ki in k:
        del out[ki]

    alpha_adj = out["alpha"].values.copy()
    alpha_var_adj = out["alpha_var"].values.copy()

    out = ds.dts.calibrate_double_ended(
        sections=sections,
        st_var=1.5,
        ast_var=1.5,
        rst_var=1.0,
        rast_var=1.0,
        method="wls",
        solver="sparse",
        fix_alpha=(alpha_adj, alpha_var_adj),
        trans_att=[50.0],
        matching_sections=[
            (
                slice(x[3 * nx_per_sec], x[4 * nx_per_sec - 1]),
                slice(x[4 * nx_per_sec], x[5 * nx_per_sec - 1]),
                True,
            )
        ],
    )

    assert_almost_equal_verbose(temp_real_celsius, out["tmpf"].values, decimal=7)
    assert_almost_equal_verbose(temp_real_celsius, out["tmpb"].values, decimal=7)
    assert_almost_equal_verbose(temp_real_celsius, out["tmpw"].values, decimal=7)
    pass


def test_double_ended_fix_alpha_gamma_matching_sections_and_one_asym_att():
    cable_len = 100.0
    nt = 3
    time = np.arange(nt)
    nx_per_sec = 2
    nx = nx_per_sec * 8
    x = np.linspace(0.0, cable_len, nx)
    ts_cold = 4.0 + np.cos(time) * 4
    ts_warm = 20.0 + -np.sin(time) * 4
    ts_ground = 6.0

    C_p = 1324  # 1/2 * E0 * v * K_+/lam_+^4
    eta_pf = np.cos(time) / 10 + 1  # eta_+ (gain factor forward channel)
    eta_pb = np.sin(time) / 10 + 1  # eta_- (gain factor backward channel)
    C_m = 5000.0
    eta_mf = np.cos(time + np.pi / 8) / 10 + 1
    eta_mb = np.sin(time + np.pi / 8) / 10 + 1
    dalpha_r = 0.005284
    dalpha_m = 0.004961
    dalpha_p = 0.005607
    gamma = 482.6
    talph_fw = 0.9
    talph_bw = 0.8

    temp_real_kelvin = np.zeros((len(x), nt)) + 273.15
    temp_real_kelvin[:nx_per_sec] += ts_cold[None]
    temp_real_kelvin[nx_per_sec : 2 * nx_per_sec] += ts_warm[None]
    temp_real_kelvin[-nx_per_sec:] += ts_cold[None]
    temp_real_kelvin[-2 * nx_per_sec : -nx_per_sec] += ts_warm[None]
    temp_real_kelvin[2 * nx_per_sec : -2 * nx_per_sec] += ts_ground
    temp_real_celsius = temp_real_kelvin - 273.15

    st = (
        eta_pf[None]
        * C_p
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_p * x[:, None])
        * np.exp(gamma / temp_real_kelvin)
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    st[4 * nx_per_sec :] *= talph_fw
    ast = (
        eta_mf[None]
        * C_m
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_m * x[:, None])
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    rst = (
        eta_pb[None]
        * C_p
        * np.exp(-dalpha_r * (-x[:, None] + cable_len))
        * np.exp(-dalpha_p * (-x[:, None] + cable_len))
        * np.exp(gamma / temp_real_kelvin)
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    rst[: 4 * nx_per_sec] *= talph_bw
    rast = (
        eta_mb[None]
        * C_m
        * np.exp(-dalpha_r * (-x[:, None] + cable_len))
        * np.exp(-dalpha_m * (-x[:, None] + cable_len))
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )

    ds = Dataset(
        {
            "TMPR": (["x", "time"], temp_real_celsius),
            "st": (["x", "time"], st),
            "ast": (["x", "time"], ast),
            "rst": (["x", "time"], rst),
            "rast": (["x", "time"], rast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "userAcquisitionTimeBW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "1"},
    )

    sections = {
        "cold": [slice(0.0, x[nx_per_sec - 1])],
        "warm": [slice(x[nx_per_sec], x[2 * nx_per_sec - 1])],
    }

    out = ds.dts.calibrate_double_ended(
        sections=sections,
        st_var=1.5,
        ast_var=1.5,
        rst_var=1.0,
        rast_var=1.0,
        method="wls",
        solver="sparse",
        trans_att=[50.0],
        matching_sections=[
            (
                slice(x[3 * nx_per_sec], x[4 * nx_per_sec - 1]),
                slice(x[4 * nx_per_sec], x[5 * nx_per_sec - 1]),
                True,
            )
        ],
    )

    # remove TA vars
    k = ["talpha_fw", "talpha_bw", "trans_att"]

    for ki in k:
        del out[ki]

    alpha_adj = out["alpha"].values.copy()
    alpha_var_adj = out["alpha_var"].values.copy()

    out = ds.dts.calibrate_double_ended(
        sections=sections,
        st_var=1.5,
        ast_var=1.5,
        rst_var=1.0,
        rast_var=1.0,
        method="wls",
        solver="sparse",
        fix_alpha=(alpha_adj, alpha_var_adj),
        fix_gamma=(gamma, 0.0),
        trans_att=[50.0],
        matching_sections=[
            (
                slice(x[3 * nx_per_sec], x[4 * nx_per_sec - 1]),
                slice(x[4 * nx_per_sec], x[5 * nx_per_sec - 1]),
                True,
            )
        ],
    )

    assert_almost_equal_verbose(temp_real_celsius, out["tmpf"].values, decimal=7)
    assert_almost_equal_verbose(temp_real_celsius, out["tmpb"].values, decimal=7)
    assert_almost_equal_verbose(temp_real_celsius, out["tmpw"].values, decimal=7)
    pass


def test_double_ended_fix_gamma_matching_sections_and_one_asym_att():
    cable_len = 100.0
    nt = 3
    time = np.arange(nt)
    nx_per_sec = 2
    nx = nx_per_sec * 8
    x = np.linspace(0.0, cable_len, nx)
    ts_cold = 4.0 + np.cos(time) * 4
    ts_warm = 20.0 + -np.sin(time) * 4
    ts_ground = 6.0

    C_p = 1324  # 1/2 * E0 * v * K_+/lam_+^4
    eta_pf = np.cos(time) / 10 + 1  # eta_+ (gain factor forward channel)
    eta_pb = np.sin(time) / 10 + 1  # eta_- (gain factor backward channel)
    C_m = 5000.0
    eta_mf = np.cos(time + np.pi / 8) / 10 + 1
    eta_mb = np.sin(time + np.pi / 8) / 10 + 1
    dalpha_r = 0.005284
    dalpha_m = 0.004961
    dalpha_p = 0.005607
    gamma = 482.6
    talph_fw = 0.9
    talph_bw = 0.8

    temp_real_kelvin = np.zeros((len(x), nt)) + 273.15
    temp_real_kelvin[:nx_per_sec] += ts_cold[None]
    temp_real_kelvin[nx_per_sec : 2 * nx_per_sec] += ts_warm[None]
    temp_real_kelvin[-nx_per_sec:] += ts_cold[None]
    temp_real_kelvin[-2 * nx_per_sec : -nx_per_sec] += ts_warm[None]
    temp_real_kelvin[2 * nx_per_sec : -2 * nx_per_sec] += ts_ground
    temp_real_celsius = temp_real_kelvin - 273.15

    st = (
        eta_pf[None]
        * C_p
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_p * x[:, None])
        * np.exp(gamma / temp_real_kelvin)
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    st[4 * nx_per_sec :] *= talph_fw
    ast = (
        eta_mf[None]
        * C_m
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_m * x[:, None])
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    rst = (
        eta_pb[None]
        * C_p
        * np.exp(-dalpha_r * (-x[:, None] + cable_len))
        * np.exp(-dalpha_p * (-x[:, None] + cable_len))
        * np.exp(gamma / temp_real_kelvin)
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )
    rst[: 4 * nx_per_sec] *= talph_bw
    rast = (
        eta_mb[None]
        * C_m
        * np.exp(-dalpha_r * (-x[:, None] + cable_len))
        * np.exp(-dalpha_m * (-x[:, None] + cable_len))
        / (np.exp(gamma / temp_real_kelvin) - 1)
    )

    ds = Dataset(
        {
            "TMPR": (["x", "time"], temp_real_celsius),
            "st": (["x", "time"], st),
            "ast": (["x", "time"], ast),
            "rst": (["x", "time"], rst),
            "rast": (["x", "time"], rast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "userAcquisitionTimeBW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "1"},
    )

    sections = {
        "cold": [slice(0.0, x[nx_per_sec - 1])],
        "warm": [slice(x[nx_per_sec], x[2 * nx_per_sec - 1])],
    }

    out = ds.dts.calibrate_double_ended(
        sections=sections,
        st_var=1.5,
        ast_var=1.5,
        rst_var=1.0,
        rast_var=1.0,
        method="wls",
        solver="sparse",
        fix_gamma=(gamma, 0.0),
        trans_att=[50.0],
        matching_sections=[
            (
                slice(x[3 * nx_per_sec], x[4 * nx_per_sec - 1]),
                slice(x[4 * nx_per_sec], x[5 * nx_per_sec - 1]),
                True,
            )
        ],
    )

    assert_almost_equal_verbose(temp_real_celsius, out["tmpf"].values, decimal=7)
    assert_almost_equal_verbose(temp_real_celsius, out["tmpb"].values, decimal=7)
    assert_almost_equal_verbose(temp_real_celsius, out["tmpw"].values, decimal=7)
    pass


@pytest.mark.skip(
    reason="Superseeded by " "test_estimate_variance_of_temperature_estimate"
)
def test_double_ended_exponential_variance_estimate_synthetic():
    import dask.array as da

    state = da.random.RandomState(0)

    stokes_m_var = 4.0
    cable_len = 100.0
    nt = 5
    time = np.arange(nt)
    x = np.linspace(0.0, cable_len, 100)
    ts_cold = np.ones(nt) * 4.0
    ts_warm = np.ones(nt) * 20.0

    C_p = 15246
    C_m = 2400.0
    dalpha_r = 0.0005284
    dalpha_m = 0.0004961
    dalpha_p = 0.0005607
    gamma = 482.6
    cold_mask = x < 0.5 * cable_len
    warm_mask = np.invert(cold_mask)  # == False
    temp_real = np.ones((len(x), nt))
    temp_real[cold_mask] *= ts_cold + 273.15
    temp_real[warm_mask] *= ts_warm + 273.15

    st = (
        C_p
        * np.exp(-(dalpha_r + dalpha_p) * x[:, None])
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    ast = (
        C_m
        * np.exp(-(dalpha_r + dalpha_m) * x[:, None])
        / (np.exp(gamma / temp_real) - 1)
    )
    rst = (
        C_p
        * np.exp(-(dalpha_r + dalpha_p) * (-x[:, None] + cable_len))
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    rast = (
        C_m
        * np.exp(-(dalpha_r + dalpha_m) * (-x[:, None] + cable_len))
        / (np.exp(gamma / temp_real) - 1)
    )

    mst_var = 1.0 * stokes_m_var
    mast_var = 1.5 * stokes_m_var
    mrst_var = 0.8 * stokes_m_var
    mrast_var = 0.5 * stokes_m_var

    st_m = st + stats.norm.rvs(size=st.shape, scale=mst_var**0.5)
    ast_m = ast + stats.norm.rvs(size=ast.shape, scale=mast_var**0.5)
    rst_m = rst + stats.norm.rvs(size=rst.shape, scale=mrst_var**0.5)
    rast_m = rast + stats.norm.rvs(size=rast.shape, scale=mrast_var**0.5)

    print("alphaint", cable_len * (dalpha_p - dalpha_m))
    print("alpha", dalpha_p - dalpha_m)
    print("C", np.log(C_p / C_m))
    print("x0", x.max())

    ds = Dataset(
        {
            # 'st':    (['x', 'time'], st),
            # 'ast':   (['x', 'time'], ast),
            # 'rst':   (['x', 'time'], rst),
            # 'rast':  (['x', 'time'], rast),
            "st": (["x", "time"], st_m),
            "ast": (["x", "time"], ast_m),
            "rst": (["x", "time"], rst_m),
            "rast": (["x", "time"], rast_m),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "userAcquisitionTimeBW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "1"},
    )

    sections = {
        "cold": [slice(0.0, 0.5 * cable_len)],
        "warm": [slice(0.5 * cable_len, cable_len)],
    }

    st_label = "st"
    ast_label = "ast"
    rst_label = "rst"
    rast_label = "rast"

    # MC variance
    ds.dts.calibrate_double_ended(
        sections=sections,
        st_label=st_label,
        ast_label=ast_label,
        rst_label=rst_label,
        rast_label=rast_label,
        st_var=mst_var,
        ast_var=mast_var,
        rst_var=mrst_var,
        rast_var=mrast_var,
        method="wls",
        solver="sparse",
    )

    ds.conf_int_double_ended(
        sections=sections,
        p_val="p_val",
        p_cov="p_cov",
        st_label=st_label,
        ast_label=ast_label,
        rst_label=rst_label,
        rast_label=rast_label,
        st_var=mst_var,
        ast_var=mast_var,
        rst_var=mrst_var,
        rast_var=mrast_var,
        conf_ints=[2.5, 50.0, 97.5],
        mc_sample_size=100,
        da_random_state=state,
    )

    # Calibrated variance
    stdsf1 = ds.ufunc_per_section(
        sections=sections, label="tmpf", func=np.std, temp_err=True, calc_per="stretch"
    )
    stdsb1 = ds.ufunc_per_section(
        sections=sections, label="tmpb", func=np.std, temp_err=True, calc_per="stretch"
    )

    # Use a single timestep to better check if the parameter uncertainties propagate
    ds1 = ds.isel(time=1)

    # Estimated VAR
    stdsf2 = ds1.ufunc_per_section(
        sections=sections,
        label="tmpf_mc_var",
        func=np.mean,
        temp_err=False,
        calc_per="stretch",
    )
    stdsb2 = ds1.ufunc_per_section(
        sections=sections,
        label="tmpb_mc_var",
        func=np.mean,
        temp_err=False,
        calc_per="stretch",
    )

    for (_, v1), (_, v2) in zip(stdsf1.items(), stdsf2.items()):
        for v1i, v2i in zip(v1, v2):
            print("Real VAR: ", v1i**2, "Estimated VAR: ", v2i)
            assert_almost_equal_verbose(v1i**2, v2i, decimal=1)

    for (_, v1), (_, v2) in zip(stdsb1.items(), stdsb2.items()):
        for v1i, v2i in zip(v1, v2):
            print("Real VAR: ", v1i**2, "Estimated VAR: ", v2i)
            assert_almost_equal_verbose(v1i**2, v2i, decimal=1)

    pass


@pytest.mark.slow  # Execution time ~2 minutes.
def test_estimate_variance_of_temperature_estimate():
    import dask.array as da

    state = da.random.RandomState(0)

    stokes_m_var = 0.1
    cable_len = 10.0
    nt = 1002
    time = np.arange(nt)
    nmc = 501
    x = np.linspace(0.0, cable_len, 64)
    ts_cold = np.ones(nt) * 4.0
    ts_warm = np.ones(nt) * 20.0

    C_p = 1524.0
    C_m = 2400.0
    dalpha_r = 0.0005284
    dalpha_m = 0.0004961
    dalpha_p = 0.0005607
    gamma = 482.6
    cold_mask = x < 0.5 * cable_len
    warm_mask = np.invert(cold_mask)  # == False
    temp_real = np.ones((len(x), nt))
    temp_real[cold_mask] *= ts_cold + 273.15
    temp_real[warm_mask] *= ts_warm + 273.15
    # alpha_int = cable_len * (dalpha_p - dalpha_m)
    # alpha = x * (dalpha_p - dalpha_m)

    st = (
        C_p
        * np.exp(-(dalpha_r + dalpha_p) * x[:, None])
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    ast = (
        C_m
        * np.exp(-(dalpha_r + dalpha_m) * x[:, None])
        / (np.exp(gamma / temp_real) - 1)
    )
    rst = (
        C_p
        * np.exp(-(dalpha_r + dalpha_p) * (-x[:, None] + cable_len))
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    rast = (
        C_m
        * np.exp(-(dalpha_r + dalpha_m) * (-x[:, None] + cable_len))
        / (np.exp(gamma / temp_real) - 1)
    )

    mst_var = 1.0 * stokes_m_var
    mast_var = 1.5 * stokes_m_var
    mrst_var = 0.8 * stokes_m_var
    mrast_var = 0.5 * stokes_m_var

    st_m = st + stats.norm.rvs(size=st.shape, scale=mst_var**0.5)
    ast_m = ast + stats.norm.rvs(size=ast.shape, scale=mast_var**0.5)
    rst_m = rst + stats.norm.rvs(size=rst.shape, scale=mrst_var**0.5)
    rast_m = rast + stats.norm.rvs(size=rast.shape, scale=mrast_var**0.5)

    print("alphaint", cable_len * (dalpha_p - dalpha_m))
    print("alpha", dalpha_p - dalpha_m)
    print("C", np.log(C_p / C_m))
    print("x0", x.max())

    ds = Dataset(
        {
            "st": (["x", "time"], st_m),
            "ast": (["x", "time"], ast_m),
            "rst": (["x", "time"], rst_m),
            "rast": (["x", "time"], rast_m),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "userAcquisitionTimeBW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "1"},
    )

    sections = {
        "cold": [slice(0.0, 0.25 * cable_len)],
        "warm": [slice(0.5 * cable_len, 0.75 * cable_len)],
    }
    # MC variance
    out = ds.dts.calibrate_double_ended(
        sections=sections,
        st_var=mst_var,
        ast_var=mast_var,
        rst_var=mrst_var,
        rast_var=mrast_var,
        # fix_gamma=(gamma, 0.),
        # fix_alpha=(alpha, 0. * alpha),
        method="wls",
        solver="stats",
    )

    out2 = ds.dts.monte_carlo_double_ended(
        result=out,
        st_var=mst_var,
        ast_var=mast_var,
        rst_var=mrst_var,
        rast_var=mrast_var,
        conf_ints=[],
        mc_sample_size=nmc,
        da_random_state=state,
        mc_remove_set_flag=False,
        reduce_memory_usage=1,
    )

    assert_almost_equal_verbose(
        (out2["r_st"] - ds["st"]).var(dim=["mc", "time"]), mst_var, decimal=2
    )
    assert_almost_equal_verbose(
        (out2["r_ast"] - ds["ast"]).var(dim=["mc", "time"]), mast_var, decimal=2
    )
    assert_almost_equal_verbose(
        (out2["r_rst"] - ds["rst"]).var(dim=["mc", "time"]), mrst_var, decimal=2
    )
    assert_almost_equal_verbose(
        (out2["r_rast"] - ds["rast"]).var(dim=["mc", "time"]), mrast_var, decimal=3
    )

    assert_almost_equal_verbose(out2["gamma_mc"].var(dim="mc"), 0.0, decimal=2)
    assert_almost_equal_verbose(out2["alpha_mc"].var(dim="mc"), 0.0, decimal=8)
    assert_almost_equal_verbose(out2["df_mc"].var(dim="mc"), out["df_var"], decimal=8)
    assert_almost_equal_verbose(out2["db_mc"].var(dim="mc"), out["db_var"], decimal=8)

    # tmpf
    temp_real2 = temp_real[:, 0] - 273.15
    actual = (out["tmpf"] - temp_real2[:, None]).var(dim="time")
    desire2 = out["tmpf_var"].mean(dim="time")

    # Validate on sections that were not used for calibration.
    assert_almost_equal_verbose(actual[16:32], desire2[16:32], decimal=2)

    # tmpb
    actual = (out["tmpb"] - temp_real2[:, None]).var(dim="time")
    desire2 = out["tmpb_var"].mean(dim="time")

    # Validate on sections that were not used for calibration.
    assert_almost_equal_verbose(actual[16:32], desire2[16:32], decimal=2)

    # tmpw
    actual = (out["tmpw"] - temp_real2[:, None]).var(dim="time")
    desire2 = out["tmpw_var"].mean(dim="time")
    assert_almost_equal_verbose(actual[16:32], desire2[16:32], decimal=3)

    pass


def test_single_ended_wls_estimate_synthetic():
    """Checks whether the coefficients are correctly defined by creating a
    synthetic measurement set, and derive the parameters from this set.
    Without variance.
    They should be the same as the parameters used to create the synthetic
    measurment set"""

    cable_len = 100.0
    nt = 50
    time = np.arange(nt)
    x = np.linspace(0.0, cable_len, 500)
    ts_cold = np.ones(nt) * 4.0
    ts_warm = np.ones(nt) * 20.0

    C_p = 15246
    C_m = 2400.0
    dalpha_r = 0.0005284
    dalpha_m = 0.0004961
    dalpha_p = 0.0005607
    gamma = 482.6
    cold_mask = x < 0.5 * cable_len
    warm_mask = np.invert(cold_mask)  # == False
    temp_real = np.ones((len(x), nt))
    temp_real[cold_mask] *= ts_cold + 273.15
    temp_real[warm_mask] *= ts_warm + 273.15

    st = (
        C_p
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_p * x[:, None])
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    ast = (
        C_m
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_m * x[:, None])
        / (np.exp(gamma / temp_real) - 1)
    )

    print("alphaint", cable_len * (dalpha_p - dalpha_m))
    print("alpha", dalpha_p - dalpha_m)
    print("C", np.log(C_p / C_m))
    print("x0", x.max())

    ds = Dataset(
        {
            "st": (["x", "time"], st),
            "ast": (["x", "time"], ast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "0"},
    )

    sections = {
        "cold": [slice(0.0, 0.5 * cable_len)],
        "warm": [slice(0.5 * cable_len, cable_len)],
    }

    # WLS
    out = ds.dts.calibrate_single_ended(
        sections=sections, st_var=1.0, ast_var=1.0, method="wls", solver="sparse"
    )

    assert_almost_equal_verbose(out["gamma"].values, gamma, decimal=6)
    assert_almost_equal_verbose(out["dalpha"].values, dalpha_p - dalpha_m, decimal=8)
    assert_almost_equal_verbose(out["tmpf"].values, temp_real - 273.15, decimal=4)

    pass


def test_single_ended_wls_fix_dalpha_synthetic():
    """Checks whether the coefficients are correctly defined by creating a
    synthetic measurement set, and derive the parameters from this set.
    Without variance.
    They should be the same as the parameters used to create the synthetic
    measurment set"""

    cable_len = 100.0
    nt = 50
    time = np.arange(nt)
    x = np.linspace(0.0, cable_len, 500)
    ts_cold = np.ones(nt) * 4.0
    ts_warm = np.ones(nt) * 20.0

    C_p = 15246
    C_m = 2400.0
    dalpha_r = 0.0005284
    dalpha_m = 0.0004961
    dalpha_p = 0.0005607
    gamma = 482.6
    cold_mask = x < 0.5 * cable_len
    warm_mask = np.invert(cold_mask)  # == False
    temp_real = np.ones((len(x), nt))
    temp_real[cold_mask] *= ts_cold + 273.15
    temp_real[warm_mask] *= ts_warm + 273.15

    st = (
        C_p
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_p * x[:, None])
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    ast = (
        C_m
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_m * x[:, None])
        / (np.exp(gamma / temp_real) - 1)
    )

    print("alphaint", cable_len * (dalpha_p - dalpha_m))
    print("alpha", dalpha_p - dalpha_m)
    print("C", np.log(C_p / C_m))
    print("x0", x.max())

    ds_ori = Dataset(
        {
            "st": (["x", "time"], st),
            "ast": (["x", "time"], ast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "0"},
    )

    sections = {
        "cold": [slice(0.0, 0.5 * cable_len)],
        "warm": [slice(0.5 * cable_len, cable_len)],
    }

    # Test fix_dalpha
    ds_dalpha = ds_ori.copy()
    out = ds_dalpha.dts.calibrate_single_ended(
        sections=sections,
        st_var=1.0,
        ast_var=1.0,
        method="wls",
        solver="sparse",
        fix_dalpha=(dalpha_p - dalpha_m, 0.0),
    )
    assert_almost_equal_verbose(out.gamma.values, gamma, decimal=12)
    assert_almost_equal_verbose(out.dalpha.values, dalpha_p - dalpha_m, decimal=14)
    assert_almost_equal_verbose(out.alpha.values, x * (dalpha_p - dalpha_m), decimal=14)
    assert_almost_equal_verbose(out.tmpf.values, temp_real - 273.15, decimal=10)

    # Test fix_alpha
    ds_alpha = ds_ori.copy()
    out = ds_alpha.dts.calibrate_single_ended(
        sections=sections,
        st_var=1.0,
        ast_var=1.0,
        method="wls",
        solver="sparse",
        fix_alpha=(x * (dalpha_p - dalpha_m), 0.0 * x),
    )

    assert_almost_equal_verbose(out.gamma.values, gamma, decimal=12)
    assert_almost_equal_verbose(out.alpha.values, x * (dalpha_p - dalpha_m), decimal=14)
    assert_almost_equal_verbose(out.tmpf.values, temp_real - 273.15, decimal=10)

    pass


def test_single_ended_wls_fix_gamma_synthetic():
    """Checks whether the coefficients are correctly defined by creating a
    synthetic measurement set, and derive the parameters from this set.
    Without variance.
    They should be the same as the parameters used to create the synthetic
    measurment set"""

    cable_len = 100.0
    nt = 50
    time = np.arange(nt)
    x = np.linspace(0.0, cable_len, 500)
    ts_cold = np.ones(nt) * 4.0
    ts_warm = np.ones(nt) * 20.0

    C_p = 15246
    C_m = 2400.0
    dalpha_r = 0.0005284
    dalpha_m = 0.0004961
    dalpha_p = 0.0005607
    gamma = 482.6
    cold_mask = x < 0.5 * cable_len
    warm_mask = np.invert(cold_mask)  # == False
    temp_real = np.ones((len(x), nt))
    temp_real[cold_mask] *= ts_cold + 273.15
    temp_real[warm_mask] *= ts_warm + 273.15

    st = (
        C_p
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_p * x[:, None])
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    ast = (
        C_m
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_m * x[:, None])
        / (np.exp(gamma / temp_real) - 1)
    )

    print("alphaint", cable_len * (dalpha_p - dalpha_m))
    print("alpha", dalpha_p - dalpha_m)
    print("C", np.log(C_p / C_m))
    print("x0", x.max())

    ds = Dataset(
        {
            "st": (["x", "time"], st),
            "ast": (["x", "time"], ast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "0"},
    )

    sections = {
        "cold": [slice(0.0, 0.5 * cable_len)],
        "warm": [slice(0.5 * cable_len, cable_len)],
    }

    # WLS
    out = ds.dts.calibrate_single_ended(
        sections=sections,
        st_var=1.0,
        ast_var=1.0,
        method="wls",
        solver="sparse",
        fix_gamma=(gamma, 0.0),
    )

    assert_almost_equal_verbose(out.gamma.values, gamma, decimal=18)
    assert_almost_equal_verbose(out.dalpha.values, dalpha_p - dalpha_m, decimal=10)
    assert_almost_equal_verbose(out.tmpf.values, temp_real - 273.15, decimal=8)

    pass


def test_single_ended_wls_fix_gamma_fix_dalpha_synthetic():
    """Checks whether the coefficients are correctly defined by creating a
    synthetic measurement set, and derive the parameters from this set.
    Without variance.
    They should be the same as the parameters used to create the synthetic
    measurment set"""

    cable_len = 100.0
    nt = 50
    time = np.arange(nt)
    x = np.linspace(0.0, cable_len, 500)
    ts_cold = np.ones(nt) * 4.0
    ts_warm = np.ones(nt) * 20.0

    C_p = 15246
    C_m = 2400.0
    dalpha_r = 0.0005284
    dalpha_m = 0.0004961
    dalpha_p = 0.0005607
    gamma = 482.6
    cold_mask = x < 0.5 * cable_len
    warm_mask = np.invert(cold_mask)  # == False
    temp_real = np.ones((len(x), nt))
    temp_real[cold_mask] *= ts_cold + 273.15
    temp_real[warm_mask] *= ts_warm + 273.15

    st = (
        C_p
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_p * x[:, None])
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    ast = (
        C_m
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_m * x[:, None])
        / (np.exp(gamma / temp_real) - 1)
    )

    print("alphaint", cable_len * (dalpha_p - dalpha_m))
    print("alpha", dalpha_p - dalpha_m)
    print("C", np.log(C_p / C_m))
    print("x0", x.max())

    ds = Dataset(
        {
            "st": (["x", "time"], st),
            "ast": (["x", "time"], ast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "0"},
    )

    sections = {
        "cold": [slice(0.0, 0.5 * cable_len)],
        "warm": [slice(0.5 * cable_len, cable_len)],
    }

    # WLS
    out = ds.dts.calibrate_single_ended(
        sections=sections,
        st_var=1.0,
        ast_var=1.0,
        method="wls",
        solver="sparse",
        fix_gamma=(gamma, 0.0),
        fix_dalpha=(dalpha_p - dalpha_m, 0.0),
    )

    assert_almost_equal_verbose(out.gamma.values, gamma, decimal=18)
    assert_almost_equal_verbose(out.dalpha.values, dalpha_p - dalpha_m, decimal=18)
    assert_almost_equal_verbose(out.tmpf.values, temp_real - 273.15, decimal=8)

    pass


def test_single_ended_trans_att_synthetic():
    """Checks whether the transient attenuation routines perform as intended,
    and calibrate to the correct temperature"""

    cable_len = 100.0
    nt = 50
    nx = 200
    time = np.arange(nt)
    x = np.linspace(0.0, cable_len, nx)
    ts_cold = np.ones(nt) * 4.0
    ts_warm = np.ones(nt) * 20.0
    ts_ambient = np.ones(nt) * 12
    ts_valid = np.ones(nt) * 16

    C_p = 15246
    C_m = 2400.0
    dalpha_r = 0.0005284
    dalpha_m = 0.0004961
    dalpha_p = 0.0005607
    gamma = 482.6
    cold_mask1 = np.logical_and(x > 0.125 * cable_len, x < 0.25 * cable_len)
    cold_mask2 = np.logical_and(x > 0.625 * cable_len, x < 0.75 * cable_len)
    warm_mask1 = np.logical_and(x > 0.75 * cable_len, x < 0.875 * cable_len)
    warm_mask2 = np.logical_and(x > 0.25 * cable_len, x < 0.375 * cable_len)
    valid_mask = np.logical_and(x > 0.40 * cable_len, x < 0.50 * cable_len)
    temp_real = np.ones((len(x), nt)) * 12 + 273.15
    temp_real[cold_mask1 + cold_mask2] = ts_cold + 273.15
    temp_real[warm_mask1 + warm_mask2] = ts_warm + 273.15
    temp_real[valid_mask] = ts_valid + 273.15

    st = (
        C_p
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_p * x[:, None])
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    ast = (
        C_m
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_m * x[:, None])
        / (np.exp(gamma / temp_real) - 1)
    )

    # Add attenuation
    tr_att = np.random.rand(nt) * 0.2 + 0.8
    st[int(x.size * 0.4) :] *= tr_att
    tr_att2 = np.random.rand(nt) * 0.2 + 0.8
    st[int(x.size * 0.6) :] *= tr_att2

    ds = Dataset(
        {
            "st": (["x", "time"], st),
            "ast": (["x", "time"], ast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
            "ambient": (["time"], ts_ambient),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "0"},
    )

    sections = {
        "ambient": [slice(0.52 * cable_len, 0.58 * cable_len)],
        "cold": [
            slice(0.125 * cable_len, 0.25 * cable_len),
            slice(0.65 * cable_len, 0.70 * cable_len),
        ],
        "warm": [slice(0.25 * cable_len, 0.375 * cable_len)],
    }

    ds_test = ds.copy(deep=True)

    # WLS
    out = ds_test.dts.calibrate_single_ended(
        sections=sections,
        st_var=1.0,
        ast_var=1.0,
        method="wls",
        trans_att=[40.0, 60.0],
        solver="sparse",
    )

    assert_almost_equal_verbose(out.gamma.values, gamma, decimal=8)
    assert_almost_equal_verbose(out.tmpf.values, temp_real - 273.15, decimal=8)
    assert_almost_equal_verbose(
        out.isel(trans_att=0).talpha_fw, -np.log(tr_att), decimal=8
    )
    assert_almost_equal_verbose(
        out.isel(trans_att=1).talpha_fw, -np.log(tr_att2), decimal=8
    )

    ds_test = ds.copy(deep=True)

    # Test fixing gamma + transient att.
    out = ds_test.dts.calibrate_single_ended(
        sections=sections,
        st_var=1.0,
        ast_var=1.0,
        method="wls",
        fix_gamma=(482.6, 0),
        trans_att=[40.0, 60.0],
        solver="sparse",
    )

    assert_almost_equal_verbose(out.gamma.values, gamma, decimal=10)
    assert_almost_equal_verbose(out.tmpf.values, temp_real - 273.15, decimal=8)
    assert_almost_equal_verbose(
        out.isel(trans_att=0).talpha_fw, -np.log(tr_att), decimal=8
    )
    assert_almost_equal_verbose(
        out.isel(trans_att=1).talpha_fw, -np.log(tr_att2), decimal=8
    )

    ds_test = ds.copy(deep=True)

    # Test fixing alpha + transient att.
    out = ds_test.dts.calibrate_single_ended(
        sections=sections,
        st_var=1.0,
        ast_var=1.0,
        method="wls",
        fix_dalpha=(6.46e-05, 0),
        trans_att=[40, 60],
        solver="sparse",
    )

    assert_almost_equal_verbose(out.gamma.values, gamma, decimal=8)
    assert_almost_equal_verbose(out.tmpf.values, temp_real - 273.15, decimal=8)
    assert_almost_equal_verbose(
        out.isel(trans_att=0).talpha_fw, -np.log(tr_att), decimal=8
    )
    assert_almost_equal_verbose(
        out.isel(trans_att=1).talpha_fw, -np.log(tr_att2), decimal=8
    )


def test_single_ended_matching_sections_synthetic():
    """Checks whether the matching sections routines perform as intended,
    and calibrate to the correct temperature"""

    cable_len = 100.0
    nt = 50
    nx = 200
    time = np.arange(nt)
    x = np.linspace(0.0, cable_len, nx)
    ts_cold = np.ones(nt) * 4.0
    ts_warm = np.ones(nt) * 20.0
    ts_ambient = np.ones(nt) * 12
    ts_valid = np.ones(nt) * 16

    C_p = 15246
    C_m = 2400.0
    dalpha_r = 0.0005284
    dalpha_m = 0.0004961
    dalpha_p = 0.0005607
    gamma = 482.6
    cold_mask1 = np.logical_and(x > 0.125 * cable_len, x < 0.25 * cable_len)
    cold_mask2 = np.logical_and(x > 0.625 * cable_len, x < 0.75 * cable_len)
    warm_mask1 = np.logical_and(x > 0.75 * cable_len, x < 0.875 * cable_len)
    warm_mask2 = np.logical_and(x > 0.25 * cable_len, x < 0.375 * cable_len)
    valid_mask = np.logical_and(x > 0.40 * cable_len, x < 0.50 * cable_len)
    temp_real = np.ones((len(x), nt)) * 12 + 273.15
    temp_real[cold_mask1 + cold_mask2] = ts_cold + 273.15
    temp_real[warm_mask1 + warm_mask2] = ts_warm + 273.15
    temp_real[valid_mask] = ts_valid + 273.15

    st = (
        C_p
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_p * x[:, None])
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    ast = (
        C_m
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_m * x[:, None])
        / (np.exp(gamma / temp_real) - 1)
    )

    # Add attenuation
    tr_att = np.random.rand(nt) * 0.2 + 0.8
    st[int(x.size * 0.4) :] *= tr_att
    tr_att2 = np.random.rand(nt) * 0.2 + 0.8
    st[int(x.size * 0.6) :] *= tr_att2

    ds = Dataset(
        {
            "st": (["x", "time"], st),
            "ast": (["x", "time"], ast),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
            "ambient": (["time"], ts_ambient),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "0"},
    )

    sections = {
        "cold": [slice(0.13 * cable_len, 0.24 * cable_len)],
        "warm": [slice(0.26 * cable_len, 0.365 * cable_len)],
    }

    matching_sections = [
        (
            slice(0.01 * cable_len, 0.09 * cable_len),
            slice(0.51 * cable_len, 0.59 * cable_len),
            True,
        ),
        (
            slice(0.01 * cable_len, 0.09 * cable_len),
            slice(0.91 * cable_len, 0.99 * cable_len),
            True,
        ),
    ]

    ds_test = ds.copy(deep=True)

    # WLS
    out = ds_test.dts.calibrate_single_ended(
        sections=sections,
        st_var=1.0,
        ast_var=1.0,
        method="wls",
        matching_sections=matching_sections,
        trans_att=[40, 60],
        solver="sparse",
    )

    assert_almost_equal_verbose(out.gamma.values, gamma, decimal=8)
    assert_almost_equal_verbose(out.tmpf.values, temp_real - 273.15, decimal=8)
    assert_almost_equal_verbose(
        out.isel(trans_att=0).talpha_fw, -np.log(tr_att), decimal=8
    )
    assert_almost_equal_verbose(
        out.isel(trans_att=1).talpha_fw, -np.log(tr_att2), decimal=8
    )

    ds_test = ds.copy(deep=True)

    # Test fixing gamma + transient att.
    out = ds_test.dts.calibrate_single_ended(
        sections=sections,
        st_var=1.0,
        ast_var=1.0,
        method="wls",
        fix_gamma=(482.6, 0),
        matching_sections=matching_sections,
        trans_att=[40, 60],
        solver="sparse",
    )

    assert_almost_equal_verbose(out.gamma.values, gamma, decimal=10)
    assert_almost_equal_verbose(out.tmpf.values, temp_real - 273.15, decimal=8)
    assert_almost_equal_verbose(
        out.isel(trans_att=0).talpha_fw, -np.log(tr_att), decimal=8
    )
    assert_almost_equal_verbose(
        out.isel(trans_att=1).talpha_fw, -np.log(tr_att2), decimal=8
    )

    ds_test = ds.copy(deep=True)

    # Test fixing dalpha + transient att.
    out = ds_test.dts.calibrate_single_ended(
        sections=sections,
        st_var=1.0,
        ast_var=1.0,
        method="wls",
        fix_dalpha=(6.46e-05, 0),
        matching_sections=matching_sections,
        trans_att=[40, 60],
        solver="sparse",
    )

    assert_almost_equal_verbose(out.gamma.values, gamma, decimal=10)
    assert_almost_equal_verbose(out.tmpf.values, temp_real - 273.15, decimal=8)
    assert_almost_equal_verbose(
        out.isel(trans_att=0).talpha_fw, -np.log(tr_att), decimal=8
    )
    assert_almost_equal_verbose(
        out.isel(trans_att=1).talpha_fw, -np.log(tr_att2), decimal=8
    )

    ds_test = ds.copy(deep=True)

    # Test fixing gamma & dalpha + transient att.
    out = ds_test.dts.calibrate_single_ended(
        sections=sections,
        st_var=1.0,
        ast_var=1.0,
        method="wls",
        fix_gamma=(482.6, 0),
        fix_dalpha=(6.46e-05, 0),
        matching_sections=matching_sections,
        trans_att=[40, 60],
        solver="sparse",
    )

    assert_almost_equal_verbose(out.gamma.values, gamma, decimal=10)
    assert_almost_equal_verbose(out.tmpf.values, temp_real - 273.15, decimal=8)
    assert_almost_equal_verbose(
        out.isel(trans_att=0).talpha_fw, -np.log(tr_att), decimal=8
    )
    assert_almost_equal_verbose(
        out.isel(trans_att=1).talpha_fw, -np.log(tr_att2), decimal=8
    )

    # Test conf. ints. for the combination of everything
    out_conf = ds_test.dts.monte_carlo_single_ended(
        result=out,
        st_var=1.0,
        ast_var=1.0,
        conf_ints=[2.5, 50.0, 97.5],
        mc_sample_size=50,
    )

    out_conf_1 = out_conf.isel(time=-1)
    # ds_test_1.tmpf
    # ds_test_1.tmpf_mc.isel(CI=0).values
    # ds_test_1.tmpf_mc.isel(CI=2).values

    assert np.all(
        np.less(out_conf_1.tmpf_mc.isel(CI=0).values, out.isel(time=-1).tmpf)
    ), "Single-ended, trans. att.; 2.5% confidence interval is incorrect"

    assert np.all(
        np.greater(out_conf_1.tmpf_mc.isel(CI=2).values, out.isel(time=-1).tmpf)
    ), "Single-ended, trans. att.; 97.5% confidence interval is incorrect"


def test_single_ended_exponential_variance_estimate_synthetic():
    """Checks whether the coefficients are correctly defined by creating a
    synthetic measurement set, and derive the parameters from this set.
    With variance.
    They should be the same as the parameters used to create the synthetic
    measurment set"""
    import dask.array as da

    state = da.random.RandomState(0)

    stokes_m_var = 40.0
    astokes_m_var = 60.0
    cable_len = 100.0
    nt = 50
    time = np.arange(nt)
    x = np.linspace(0.0, cable_len, 500)
    ts_cold = np.ones(nt) * 4.0
    ts_warm = np.ones(nt) * 20.0

    C_p = 15246
    C_m = 2400.0
    dalpha_r = 0.0005284
    dalpha_m = 0.0004961
    dalpha_p = 0.0005607
    gamma = 482.6
    cold_mask = x < 0.5 * cable_len
    warm_mask = np.invert(cold_mask)  # == False
    temp_real = np.ones((len(x), nt))
    temp_real[cold_mask] *= ts_cold + 273.15
    temp_real[warm_mask] *= ts_warm + 273.15

    st = (
        C_p
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_p * x[:, None])
        * np.exp(gamma / temp_real)
        / (np.exp(gamma / temp_real) - 1)
    )
    ast = (
        C_m
        * np.exp(-dalpha_r * x[:, None])
        * np.exp(-dalpha_m * x[:, None])
        / (np.exp(gamma / temp_real) - 1)
    )
    st_m = st + stats.norm.rvs(size=st.shape, scale=stokes_m_var**0.5)
    ast_m = ast + stats.norm.rvs(size=ast.shape, scale=astokes_m_var**0.5)

    # print('alphaint', cable_len * (dalpha_p - dalpha_m))
    # print('alpha', dalpha_p - dalpha_m)
    # print('C', np.log(C_p / C_m))
    # print('x0', x.max())

    ds = Dataset(
        {
            # 'st':                    (['x', 'time'], st),
            # 'ast':                   (['x', 'time'], ast),
            "st": (["x", "time"], st_m),
            "ast": (["x", "time"], ast_m),
            "userAcquisitionTimeFW": (["time"], np.ones(nt)),
            "cold": (["time"], ts_cold),
            "warm": (["time"], ts_warm),
        },
        coords={"x": x, "time": time},
        attrs={"isDoubleEnded": "0"},
    )

    sections = {
        "cold": [slice(0.0, 0.5 * cable_len)],
        "warm": [slice(0.5 * cable_len, cable_len)],
    }

    mst_var, _ = variance_stokes_exponential(
        ds["st"],
        sections,
        ds["userAcquisitionTimeFW"],
        use_statsmodels=False,
        suppress_info=False,
        reshape_residuals=True,
    )
    mast_var, _ = variance_stokes_exponential(
        ds["ast"],
        sections,
        ds["userAcquisitionTimeFW"],
        use_statsmodels=False,
        suppress_info=False,
        reshape_residuals=True,
    )

    # MC variqnce
    out = ds.dts.calibrate_single_ended(
        sections=sections,
        st_var=mst_var,
        ast_var=mast_var,
        method="wls",
        solver="sparse",
    )
    ds2 = ds.copy()
    ds2.update(out)

    # Calibrated variance
    stdsf1 = ds2.dts.ufunc_per_section(
        sections=sections,
        label="tmpf",
        func=np.var,
        temp_err=True,
        calc_per="stretch",
        ddof=1,
    )
    out_ci = ds2.dts.monte_carlo_single_ended(
        result=out,
        st_var=mst_var,
        ast_var=mast_var,
        conf_ints=[2.5, 50.0, 97.5],
        mc_sample_size=50,
        da_random_state=state,
        mc_remove_set_flag=False,
    )
    ds2.update(out_ci)

    # Use a single timestep to better check if the parameter uncertainties
    # propagate
    # Estimated VAR
    stdsf2 = ds2.isel(time=1).dts.ufunc_per_section(
        sections=sections,
        label="tmpf_mc_var",
        func=np.mean,
        temp_err=False,
        calc_per="stretch",
    )

    for (_, v1), (_, v2) in zip(stdsf1.items(), stdsf2.items()):
        for v1i, v2i in zip(v1, v2):
            v2i_c = float(v2i)
            print("Real VAR: ", v1i, "Estimated VAR: ", v2i_c)
            assert_almost_equal_verbose(v1i, v2i_c, decimal=1)

    pass
    print("hoi")


def test_calibrate_wls_solver_procedures():
    x = np.linspace(0, 10, 25 * 4)
    np.random.shuffle(x)

    X = x.reshape((25, 4))
    beta = np.array([1, 0.1, 10, 5])
    beta_w = np.concatenate((np.ones(10), np.ones(15) * 1.0))
    beta_0 = np.array([1, 1, 1, 1])
    y = np.dot(X, beta)
    y_meas = y + np.random.normal(size=y.size)

    # first check unweighted convergence
    beta_numpy = np.linalg.lstsq(X, y, rcond=None)[0]
    np.testing.assert_array_almost_equal(beta, beta_numpy, decimal=8)

    ps_sol, ps_var = wls_stats(X, y, w=1, calc_cov=0)
    p_sol, p_var = wls_sparse(X, y, w=1, calc_cov=0, x0=beta_0)

    np.testing.assert_array_almost_equal(beta, ps_sol, decimal=8)
    np.testing.assert_array_almost_equal(beta, p_sol, decimal=8)

    # now with weights
    dec = 8
    ps_sol, ps_var, ps_cov = wls_stats(X, y_meas, w=beta_w, calc_cov=True, x0=beta_0)
    p_sol, p_var, p_cov = wls_sparse(X, y_meas, w=beta_w, calc_cov=True, x0=beta_0)

    np.testing.assert_array_almost_equal(p_sol, ps_sol, decimal=dec)
    np.testing.assert_array_almost_equal(p_var, ps_var, decimal=dec)
    np.testing.assert_array_almost_equal(p_cov, ps_cov, decimal=dec)

    # Test array sparse
    Xsp = sp.coo_matrix(X)
    psp_sol, psp_var, psp_cov = wls_stats(Xsp, y_meas, w=beta_w, calc_cov=True)

    np.testing.assert_array_almost_equal(p_sol, psp_sol, decimal=dec)
    np.testing.assert_array_almost_equal(p_var, psp_var, decimal=dec)
    np.testing.assert_array_almost_equal(p_cov, psp_cov, decimal=dec)

    pass



"""Common utilities for reading input files."""
from contextlib import contextmanager

import pandas as pd

# Returns a dictionary with the attributes to the dimensions.
#  The keys refer to the naming used in the raw files.
# TODO: attrs for st_var and use it in parse_st_var function
_dim_attrs = {
    ("x", "distance"): dict(
        name="distance",
        description="Length along fiber",
        long_description="Starting at connector of forward channel",
        units="m",
    ),
    ("tmp", "temperature"): dict(
        name="tmp", description="Temperature calibrated by device", units=r"$^\circ$C"
    ),
    ("st",): dict(name="st", description="Stokes intensity", units="-"),
    ("ast",): dict(name="ast", description="anti-Stokes intensity", units="-"),
    ("rst",): dict(name="rst", description="reverse Stokes intensity", units="-"),
    ("rast",): dict(
        name="rast", description="reverse anti-Stokes intensity", units="-"
    ),
    ("tmpf",): dict(
        name="tmpf",
        description="Temperature estimated using the Stokes and anti-Stokes from the Forward channel",
        units=r"$^\circ$C",
    ),
    ("tmpb",): dict(
        name="tmpb",
        description="Temperature estimated using the Stokes and anti-Stokes from the Backward channel",
        units=r"$^\circ$C",
    ),
    ("tmpw",): dict(
        name="tmpw",
        description="Temperature estimated using the Stokes and anti-Stokes from both the Forward and Backward channel.",
        units=r"$^\circ$C",
    ),
    ("tmpf_var",): dict(
        name="tmpf_var",
        description="Uncertainty variance in tmpf estimated with linear-error propagation",
        units=r"$^\circ$C$^2$",
    ),
    ("tmpb_var",): dict(
        name="tmpb_var",
        description="Uncertainty variance in tmpb estimated with linear-error propagation",
        units=r"$^\circ$C$^2$",
    ),
    ("tmpw_var",): dict(
        name="tmpw_var",
        description="Uncertainty variance in tmpw estimated with linear-error propagation",
        units=r"$^\circ$C$^2$",
    ),
    ("tmpw_var_lower",): dict(
        name="tmpw_var_lower",
        description="Lower bound of uncertainty variance in tmpw estimated with linear-error propagation. "
        "Excludes the parameter uncertainties.",
        units=r"$^\circ$C$^2$",
    ),
    ("tmpw_var_approx",): dict(
        name="tmpw_var_upper",
        description="Upper bound of uncertainty variance in tmpw estimated with linear-error propagation. "
        "Excludes the correlation between tmpf and tmpb caused by alpha.",
        units=r"$^\circ$C$^2$",
    ),
    ("tmpf_mc_var",): dict(
        name="tmpf_mc_var",
        description="Uncertainty variance in tmpf estimated with Monte Carlo",
        units=r"$^\circ$C$^2$",
    ),
    ("tmpb_mc_var",): dict(
        name="tmpb_mc_var",
        description="Uncertainty variance in tmpb estimated with Monte Carlo",
        units=r"$^\circ$C$^2$",
    ),
    ("tmpw_mc_var",): dict(
        name="tmpw_mc_var",
        description="Uncertainty variance in tmpw estimated with Monte Carlo",
        units=r"$^\circ$C$^2$",
    ),
    ("acquisitionTime",): dict(
        name="acquisitionTime",
        description="Measurement duration of forward channel",
        long_description="Actual measurement duration of forward " "channel",
        units="seconds",
    ),
    ("userAcquisitionTimeFW",): dict(
        name="userAcquisitionTimeFW",
        description="Measurement duration of forward channel",
        long_description="Desired measurement duration of forward " "channel",
        units="seconds",
    ),
    ("userAcquisitionTimeBW",): dict(
        name="userAcquisitionTimeBW",
        description="Measurement duration of backward channel",
        long_description="Desired measurement duration of backward " "channel",
        units="seconds",
    ),
    ("trans_att",): dict(
        name="Locations introducing transient directional differential " "attenuation",
        description="Locations along the x-dimension introducing transient "
        "directional differential attenuation",
        long_description="Connectors introduce additional differential "
        "attenuation that is different for the forward "
        "and backward direction, and varies over time.",
        units="m",
    ),
}

# Because variations in the names exist between the different file formats. The
#   tuple as key contains the possible keys, which is expanded below.
dim_attrs = {k: v for kl, v in _dim_attrs.items() for k in kl}


@contextmanager
def open_file(path, **kwargs):
    if isinstance(path, tuple):
        # print('\nabout to open zipfile', path[0], '. from', path[1])
        # assert isinstance(path[1], zip)
        with path[1].open(path[0], **kwargs) as the_file:
            yield the_file
    else:
        with open(path, **kwargs) as the_file:
            yield the_file


def get_xml_namespace(element):
    """Parameters
    ----------
    element

    Returns:
    --------

    """
    import re

    m = re.match("\\{.*\\}", element.tag)
    return m.group(0) if m else ""


def coords_time(
    maxTimeIndex,
    timezone_input_files=None,
    timezone_netcdf="UTC",
    dtFW=None,
    dtBW=None,
    double_ended_flag=False,
):
    """Prepares the time coordinates for the construction of DataStore
    instances with metadata.

    Parameters
    ----------
    maxTimeIndex : array-like (1-dimensional)
        Is an array with 'datetime64[ns]' timestamps of the end of the
        forward channel. If single ended this is the end of the measurement.
        If double ended this is halfway the double ended measurement.
    timezone_input_files : string, pytz.timezone, dateutil.tz.tzfile or None
        A string of a timezone that is understood by pandas of maxTimeIndex.
        If None, it is assumed that the input files are already timezone aware
    timezone_netcdf : string, pytz.timezone, dateutil.tz.tzfile or None
        A string of a timezone that is understood by pandas to write the
        netCDF to. Using UTC as default, according to CF conventions.
    dtFW : array-like (1-dimensional) of float
        The acquisition time of the Forward channel in seconds
    dtBW : array-like (1-dimensional) of float
        The acquisition time of the Backward channel in seconds
    double_ended_flag : bool
        A flag whether the measurement is double ended

    Returns:
    --------

    """
    time_attrs = {
        "time": {
            "description": "time halfway the measurement",
            "timezone": str(timezone_netcdf),
        },
        "timestart": {
            "description": "time start of the measurement",
            "timezone": str(timezone_netcdf),
        },
        "timeend": {
            "description": "time end of the measurement",
            "timezone": str(timezone_netcdf),
        },
        "timeFW": {
            "description": "time halfway the forward channel measurement",
            "timezone": str(timezone_netcdf),
        },
        "timeFWstart": {
            "description": "time start of the forward channel measurement",
            "timezone": str(timezone_netcdf),
        },
        "timeFWend": {
            "description": "time end of the forward channel measurement",
            "timezone": str(timezone_netcdf),
        },
        "timeBW": {
            "description": "time halfway the backward channel measurement",
            "timezone": str(timezone_netcdf),
        },
        "timeBWstart": {
            "description": "time start of the backward channel measurement",
            "timezone": str(timezone_netcdf),
        },
        "timeBWend": {
            "description": "time end of the backward channel measurement",
            "timezone": str(timezone_netcdf),
        },
    }

    if not double_ended_flag:
        # single ended measurement
        dt1 = dtFW.astype("timedelta64[s]")

        # start of the forward measurement
        index_time_FWstart = maxTimeIndex - dt1

        # end of the forward measurement
        index_time_FWend = maxTimeIndex

        # center of forward measurement
        index_time_FWmean = maxTimeIndex - dt1 / 2

        coords_zip = [
            ("timestart", index_time_FWstart),
            ("timeend", index_time_FWend),
            ("time", index_time_FWmean),
        ]

    else:
        # double ended measurement
        dt1 = dtFW.astype("timedelta64[s]")
        dt2 = dtBW.astype("timedelta64[s]")

        # start of the forward measurement
        index_time_FWstart = maxTimeIndex - dt1

        # end of the forward measurement
        index_time_FWend = maxTimeIndex

        # center of forward measurement
        index_time_FWmean = maxTimeIndex - dt1 / 2

        # start of the backward measurement
        index_time_BWstart = index_time_FWend.copy()

        # end of the backward measurement
        index_time_BWend = maxTimeIndex + dt2

        # center of backward measurement
        index_time_BWmean = maxTimeIndex + dt2 / 2

        coords_zip = [
            ("timeFWstart", index_time_FWstart),
            ("timeFWend", index_time_FWend),
            ("timeFW", index_time_FWmean),
            ("timeBWstart", index_time_BWstart),
            ("timeBWend", index_time_BWend),
            ("timeBW", index_time_BWmean),
            ("timestart", index_time_FWstart),
            ("timeend", index_time_BWend),
            ("time", index_time_FWend),
        ]

    if timezone_input_files is not None:
        coords = {
            k: (
                "time",
                pd.DatetimeIndex(v)
                .tz_localize(tz=timezone_input_files)
                .tz_convert(timezone_netcdf)
                .tz_localize(None)
                .astype("datetime64[ns]"),
                time_attrs[k],
            )
            for k, v in coords_zip
        }
    else:
        coords = {
            k: (
                "time",
                pd.DatetimeIndex(v)
                .tz_convert(timezone_netcdf)
                .tz_localize(None)
                .astype("datetime64[ns]"),
                time_attrs[k],
            )
            for k, v in coords_zip
        }

    # The units are already stored in the dtype
    coords["acquisitiontimeFW"] = (
        "time",
        dt1,
        {"description": "Acquisition time of the forward measurement"},
    )

    if double_ended_flag:
        # The units are already stored in the dtype
        coords["acquisitiontimeBW"] = (
            "time",
            dt2,
            {"description": "Acquisition time of the backward measurement"},
        )

    return coords


def ziphandle_to_filepathlist(fh=None, extension=None):
    fnl_ = sorted(fh.namelist())

    fnl = []
    for name in fnl_:
        if name[:1] == "_":
            # private POSIX
            continue

        if fh.getinfo(name).is_dir():
            continue

        if not name.endswith(extension.strip("*")):
            continue

        fnl.append((name, fh))

    return fnl
